/*
Copyright (c) 2009, Miroslav Kulich
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the Player Project nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/** This file implements the Stage controller which connects to Player server and displays data provided by the server in the Stage window. Possible applications:
 *  - real robot monitoring in case Player runs on the robot
 *  - monitoring of several (both real and simulated) robots on one window
 *  - supervising student's robot by a teacher
 * and more...
 * 
 * Example of controller configuration (in world file):
 * fancypioneer2dx (		 
 * name "r0"
 * ctrl "viewer ( host 'mycomp.com'  port 3333 requires ['laser:0' 'position2d:0' 'sonar:0']) "
 * sicklaser()
 * ranger()
 * )
 * In this configuration, the Stage (as standalone app) connects to computer  mycom.com on port 3333 and displays data from
 * its laser:0 device as it is generated by sicklaser and data from sonar:0 as ranger. Moreover, position of
 * the displayed robot is position2d:0. 
 * 
 * Format:
 * ctrl "viewer ( host '<HOST>' port <PORT> requires ' <DEVICE1> <DEVICE2> ... <DEVICEn>' )"
 * <HOST> name of Player host computer
 * <PORT> port on whict Player communicates
 * <DEVICEi> The device address(es) through which the driver can be accessed.
 */

#include "stage.hh"
#include "libstage/worldfile.hh"
#include <libplayerc++/playerc++.h>
#include <pthread.h>
#include <iostream>
#include <fstream>
#include "viewer.hh"

using namespace Stg;
using namespace PlayerCc;


std::map< uint16_t, ProxyConfiguration> proxy_map;
bool RobotConnector::done = false;
std::vector<pthread_t> RobotConnector::threads;

// Template used to created named wrappers for model constructors 
template <class T>
ClientProxy* Creator (PlayerClient *aPc, uint32_t aIndex) {
   return new T( aPc, aIndex );
};

static void Register( uint16_t interf, constructor_t func, const std::string& model, stg_model_callback_t callback, init_func_t initfunc) {
   proxy_map[ interf ] = ProxyConfiguration(func, model, callback, initfunc);  
}



void registerProxies() {
   static bool done = false;
   if (!done) {
      Register(PLAYER_POSITION2D_CODE,  Creator<Position2dProxy>, "position", RobotConnector::PositionUpdate, 0);
      Register(PLAYER_RANGER_CODE,  Creator<RangerProxy>, "ranger", RobotConnector::RangerUpdate, RobotConnector::RangerInit);
      Register(PLAYER_SONAR_CODE,  Creator<SonarProxy>, "sonar", RobotConnector::SonarUpdate, RobotConnector::SonarInit);
      Register(PLAYER_LASER_CODE,  Creator<LaserProxy>, "laser", RobotConnector::LaserUpdate, RobotConnector::LaserInit);
   //   Register(PLAYER_GRAPHICS2D_CODE,  Creator<Graphics2dProxy>, "graphics2d");
      done = true;
   }
}





// Called when the Stage exits 
void exitCallback() {
   RobotConnector::done = true;
   void *status;
   FOR_EACH(id,RobotConnector::threads) {
      pthread_join(*id,&status);
   }
   RobotConnector::threads.clear();
   //TODO descructs all robots   
   //   delete robot;
   //FOR_EACH(pair,proxies) {
   //  delete (*pair);
}

void* RobotConnector::readerThread(void* robotConnector) {
   RobotConnector* rc = (RobotConnector*) robotConnector; 
   uint16_t index;
   while(!done) {
      rc->robot->Read();
      pthread_mutex_lock(&rc->mutex);
      FOR_EACH(pair,rc->proxies) {
         index = (*pair)->index;
         switch ((*pair)->proxy->GetInterface()) {
         case  PLAYER_POSITION2D_CODE: {
            Position2dProxy *pp = (Position2dProxy*) ((*pair)->proxy);
            rc->state[index].position = Pose(pp->GetXPos(),pp->GetYPos(),0,pp->GetYaw());
         }
            break;
         case PLAYER_LASER_CODE: {
            LaserProxy *pp = (LaserProxy *) (*pair)->proxy;
            rc->state[index].laser.resize(pp->GetCount());
            for(unsigned int i=0; i<pp->GetCount();i++) {
               rc->state[index].laser[i].range = pp->GetRange(i);
               rc->state[index].laser[i].reflectance = pp->GetIntensity(i);
            }
         }
            break;
         case PLAYER_SONAR_CODE: {
            SonarProxy *pp = (SonarProxy *) (*pair)->proxy;
            rc->state[index].sonar.resize(pp->GetCount());
            for(unsigned int i=0; i<pp->GetCount();i++) {
               rc->state[index].sonar[i] = pp->GetScan(i);
            }
         }
         case PLAYER_RANGER_CODE: {
            RangerProxy *pp = (RangerProxy *) (*pair)->proxy;
            rc->state[index].ranger.resize(pp->GetRangeCount());
            pp->RequestConfigure ();
            player_pose3d_t p = pp->GetDevicePose();
                   
            double angle = pp->GetMaxAngle();
            double res = /*p.pyaw + */ pp->GetAngularRes();
            for(unsigned int i=0; i<pp->GetRangeCount();i++) {
               rc->state[index].ranger[i].range = pp->GetRange(i);
               rc->state[index].ranger[i].pose = Pose(0,0,0,angle);//Pose(p.px,p.py,p.pz,angle);
               rc->state[index].ranger[i].fov = pp->GetRangeRes()*4;
               angle -= res;
            }
          }
          break;
         default:
            PRINT_WARN2("proxy  %s(%d) not supported.\n",(*pair)->proxy->GetInterfaceStr().c_str(),(*pair)->proxy->GetInterface());
         }
         pthread_mutex_unlock(&rc->mutex);
      }
   }
   return 0;
}




void RobotConnector::loadParams(std::string args) {
   std::string filename = "tmp";
   std::ofstream out;
   out.open (filename.c_str());
   unsigned int p; 
   
   // replace all  ' with " 
   while ( (p=args.find('\'')) != std::string::npos) {
      args[p] = '\"';
   }
   out << args;
   out.close();
   
   Stg::Worldfile wf;
   wf.Load(filename.c_str());
   remove(filename.c_str());
   computer = wf.ReadString(1,"host","localhost");
   port = wf.ReadInt(1,"port",6665);
   printf("\nConnecting to  %s:%d.\n",computer.c_str(),port);
   const char* val;
   CProperty* property = wf.GetProperty(1, "requires");
   if (property != 0) {
      for(unsigned int i=0;i<property->values.size();i++) {
         val =  wf.GetPropertyValue(property, i);
         interfaces.insert(std::string(val));
         printf("Interface %d: [%s]\n",i,val);
      }
   }
}

 //TODO  find child also with a specified index
Model* RobotConnector::getChildWithType(Model* mod,std::string type)
{
   Model *result = mod->GetModelType() == type ? mod : 0;
   FOR_EACH(it, mod->GetChildren()) {
      if ((*it)->GetModelType() == type && used[*it]==false) { 
         result = *it;
      }
   }
//  printf("Model %s (%s)\n",  mod->GetModelType().c_str(),mod->Token());
   return result;
}


void RobotConnector::LaserInit(Model *mod,ClientProxy* proxy) {
   ModelLaser *m = (ModelLaser*) mod;
   ((LaserProxy *)proxy)->RequestGeom();
   player_pose3d_t  pose = ((LaserProxy *)proxy)->GetPose();
   Pose pp = m->GetPose();
   pp.x = pose.px;
   pp.y = pose.py;
   pp.a = pose.pyaw;   
   m->SetPose(pp);
}

void RobotConnector::RangerInit(Model *mod,ClientProxy* proxy) {
   ModelRanger *m = (ModelRanger*) mod;
   ((LaserProxy *)proxy)->RequestGeom();
   player_pose3d_t  pose = ((RangerProxy *)proxy)->GetDevicePose();
   Pose pp = m->GetPose();
   pp.x = pose.px;
   pp.y = pose.py;
   pp.a = pose.pyaw;
   m->SetPose(pp);
}

void RobotConnector::SonarInit(Model *mod,ClientProxy* proxy) {
   //   return;
   player_pose3d_t  pose;
   Pose pp;
   ModelRanger *m = (ModelRanger*) mod;
   SonarProxy *prx = (SonarProxy *)proxy;
   prx->RequestGeom();
   printf("count: %d\n",prx->GetPoseCount());
   m->sensors.resize(prx->GetPoseCount());
   for(unsigned int i=0;i<prx->GetPoseCount();i++) {
      pose = prx->GetPose(i);
      m->sensors[i].pose.x = pose.px;
      m->sensors[i].pose.y = pose.py;
      m->sensors[i].pose.a = pose.pyaw;
      //      printf("%d %f %f %f\n",i, m->sensors[i].pose.x,m->sensors[i].pose.y,m->sensors[i].pose.a);
   }
}



int RobotConnector::PositionUpdate( Model* mod, void* pair )
{
   uint16_t index  = ((UpdateConfiguration*) pair)->index;
   RobotConnector *rc = ((UpdateConfiguration*) pair)->connector;
   ModelPosition *pos = (ModelPosition*)mod;
   pthread_mutex_lock(&rc->mutex);
   pos->SetPose(rc->state[index].position);
   pthread_mutex_unlock(&rc->mutex);
   return 0; // run again
}

int RobotConnector::LaserUpdate( Model* mod, void* pair )
{
   uint16_t index  = ((UpdateConfiguration*) pair)->index;
   RobotConnector *rc = ((UpdateConfiguration*) pair)->connector;
   ModelLaser *laser = (ModelLaser*)mod;
   pthread_mutex_lock(&rc->mutex);
   laser->SetSamples(rc->state[index].laser);
   pthread_mutex_unlock(&rc->mutex);
   
   return 0; // run again
}

int RobotConnector::RangerUpdate( Model* mod, void* pair )
{
   uint16_t index  = ((UpdateConfiguration*) pair)->index;
   RobotConnector *rc = ((UpdateConfiguration*) pair)->connector;
   ModelRanger *ranger = (ModelRanger*)mod;
   pthread_mutex_lock(&rc->mutex);
   int i=0;
   ranger->sensors.resize(rc->state[index].ranger.size());
   FOR_EACH(it,rc->state[index].ranger) {
      ranger->sensors[i].range = it->range;
      ranger->sensors[i].pose = it->pose;
      ranger->sensors[i].fov = it->fov;
      i++;
   }

   pthread_mutex_unlock(&rc->mutex);

   return 0; // run again
}

int RobotConnector::SonarUpdate( Model* mod, void* pair )
{
   uint16_t index  = ((UpdateConfiguration*) pair)->index;
   RobotConnector *rc = ((UpdateConfiguration*) pair)->connector;
   ModelRanger *ranger = (ModelRanger*)mod;
   
   pthread_mutex_lock(&rc->mutex);
   int i=0;
   FOR_EACH(it,rc->state[index].sonar) {
      ranger->sensors[i].range = (*it);
      i++;
   }
   pthread_mutex_unlock(&rc->mutex);
   
   return 0; // run again
}


RobotConnector::RobotConnector(Model* mod, CtrlArgs* args) {
   loadParams(args->worldfile);
   FOR_EACH(it, mod->GetChildren()) {
      used[*it] = false;
   }
   
   
   
   robot = new PlayerClient(computer,port);
   robot->RequestDeviceList();
   
   std::list< playerc_device_info_t > devices = robot->GetDeviceList();
   UpdateConfiguration *pp;
   puts("DEVICES");
   FOR_EACH( iter, devices ) {
      std::map< uint16_t, ProxyConfiguration>::iterator it = proxy_map.find(iter->addr.interf);
      std::stringstream out;
      out << robot->LookupName(iter->addr.interf) << ":" << iter->addr.index;
      std::string fullName  = out.str();
      if (interfaces.find(out.str())!=interfaces.end()) {
         if (it != proxy_map.end()) {
            Model *m = getChildWithType(mod,it->second.modelType); //model
            if (m != 0) {
               used[m] = true;
               ClientProxy *proxy  = (*it->second.constructor)( robot, iter->addr.index );
               if (it->second.initfunc != 0) {
                  (*it->second.initfunc)(m,proxy);
               }
               proxies.insert(proxies.end(),pp = new UpdateConfiguration(proxy,iter->addr.index,this));
               m->AddUpdateCallback(it->second.callback, (void*) (pp));
               m->Subscribe();
               printf ("device %s:%d -> %s\n",robot->LookupName(iter->addr.interf).c_str(),iter->addr.index, m->Token());
            } else {
               PRINT_WARN2( "device %s:%d:   model is not created.", robot->LookupName(iter->addr.interf).c_str(),iter->addr.index );
            }
         } else {
            PRINT_WARN1( "device %s  is not supported.", robot->LookupName(iter->addr.interf).c_str() );
         }
      }
   }
   
}


void RobotConnector::Run() {
   pthread_mutex_init(&mutex, 0);
   pthread_t id;
   pthread_create(&id, 0, readerThread, (void *)this);
   threads.push_back(id);
}

// Stage calls this when the model starts up
extern "C" int Init( Model* mod, CtrlArgs* args )
{
   registerProxies();
   atexit(exitCallback);
// local arguments
   printf( "\nViewer controller initialised with:\n"
           "\tworldfile string \"%s\"\n" 
           "\tcmdline string \"%s\"",
           args->worldfile.c_str(),
           args->cmdline.c_str() );
   RobotConnector *rc = new RobotConnector(mod,args);
   rc->Run();
   return 0; //ok
}


int main(void)
{
   printf( "\nViewer controller initialised with:\n"
           "\tworldfile string \"%s\"\n" 
           "\tcmdline string \"%s\"",
           "AAA",
           "BBB" );
   registerProxies();
}
