
/** canvas.cc
    Implement the main world viewing area in FLTK and OpenGL. 

    Authors: 
    Richard Vaughan (vaughan@sfu.ca)
    Alex Couture-Beil (asc17@sfu.ca)
    Jeremy Asher (jra11@sfu.ca)

    $Id: canvas.cc 8330 2009-10-27 19:13:58Z rtv $
*/

#include "stage.hh"
#include "canvas.hh"
#include "worldfile.hh"
#include "texture_manager.hh"
#include "replace.h"

#include <string>
#include <sstream>
#include <png.h>
#include <algorithm>
#include <functional>      // For greater<int>( )
#include <vlc/vlc.h> // for vlc-based video player

#include "region.hh"
#include "file_manager.hh"
#include "options_dlg.hh"
#include "image.hh"

#define VIDEOWIDTH 1296
#define VIDEOHEIGHT 960


using namespace Stg;

static  const int checkImageWidth = 2;
static  const int	checkImageHeight = 2;
static  GLubyte checkImage[checkImageHeight][checkImageWidth][4];
static bool blur = true;

static bool init_done = false;

//GLuint glowTex;
GLuint checkTex;

void setTsaiCameraView(double rotX,  /* camera rotation  */
   double rotY,
   double rotZ,
   double t_x,                /* camera translation */
   double t_y,
   double t_z,
   double focalLength,           /* effetive focal length, or image
                           plane to projective center distance, */
   double scaleX,                /* Tsai sx parameter
                        uncertainty scale factor for x, due to
                        TV camera scanning and acquisitino timing
                        error.*/
   double heightImage,              /* window in pixels widthxheight */
   double widthImage,
   double heightWindow,             /* window in pixels widthxheight */
   double widthWindow,
   double dpx,
   double dpy,
   double cXImage,               /* "computer image coordinate for */
   double cYImage)               /* the origin in the image plane */
#if 0
   double *cXError,           /* the returned error in the X and Y */
   double *cYError)
#endif
{

   double arotZ ; /* angle rotation in Z */
   double arotY ; /* angle rotation in Y */
   double arotX ; /* angle rotation in X */
   double zNear ; /* distance to near clipping plane of frustum */
   double zFar ; /* distance to far clipping plane frustum */
   double fprime ; /* focal length in OpenGL Perspective matrix */
   double aspect ; /* aspect X/Y in OpenGL Perspective matrix */
   double fovy ; /* field of view angle in vertical direction
               for OpenGL Perspective matrix */
      /* matrix is transposed, for OpenGL though looks same */
      /* used to reflect the w coordinate */
   float reflect[16] = {1, 0, 0, 0,
                         0, -1, 0, 0,
                   0, 0, -1, 0,
                   0, 0, 0, 1};

   double cXWindow;  /* the center in X coordinates of Window/Viewport */
   double cYWindow;  /* the center in Y coordinates of Window/Viewport */
   int   xWindow;    /* the lower left corner of the window */
   int yWindow;
   double xWindowD;  /* xWindow in double form before rounding */
   double yWindowD;  /* yWindow in double form before rounding */

      /* place Camera's intrinsic parameters upon the projection matrix */
      /* solve for field of view */
   fprime = 2.0/dpy * focalLength / ( heightImage);
   fovy = 2.0*atan(1.0/fprime)*180.0/M_PI;

      /* compute the aspect ratio */
   aspect = widthImage/heightImage/scaleX*dpx/dpy;

      /* set zNear and zFar to some heuristic to include the
      whole volume */
      /* 1000 and 3000 are initial guesses */
   zNear = 1.0;
   zFar  = 1000000.0;



      /* Set Perspective projection matrix */
   glMatrixMode (GL_PROJECTION);
   glLoadIdentity ();
   assert(zNear > 0.0 && zFar > 0.0);
//   printf("fprime %lf fovy %lf aspect %lf zNear %lf zFar %lf\n",fprime, fovy, aspect, zNear, zFar);

   gluPerspective(fovy, aspect, zNear, zFar);


      /* need to set viewport to incorporate Cx Cy */

      /* initialize matrix mode to affect model matrix */

   cXWindow = cXImage*widthWindow/widthImage;
   cYWindow = cYImage*heightWindow/heightImage;
   xWindowD = cXWindow - widthWindow/2.0 ; /* round */
      /* y is flipped so take negative */
   yWindowD = - ( cYWindow - heightWindow/2.0 ) ; /* round */
   xWindow = (xWindowD< 0.0) ? (int)(xWindowD - 0.5):
            (int)(xWindowD + 0.5); /* round */
   yWindow = (yWindowD < 0.0) ? (int)(yWindowD - 0.5):
            (int)(yWindowD + 0.5); /* round */



GLdouble matrix[16] = {
   1, 0, 0, 0,
   0, 1, 0, 0,
   0, 0, -1, -1,
   0, 0, -2, 0};

matrix[0] = fprime/aspect;
matrix[5] = fprime;
matrix[8] = -2*xWindow/widthWindow;
matrix[9] = -2*yWindow/heightWindow;
//printf("scaleX %lf focalLength %lf dpx  %lf dpy %lf\n\n", scaleX,focalLength,dpx, dpy);
//print("matrix",matrix);

glLoadMatrixd( matrix );

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity ();
   glMultMatrixf(reflect);

      /* first rotate OpenGL camera 180 degrees to correspond to
         Tsai's camera */
   arotZ = rotZ * 180.0 / M_PI;
   arotY = rotY * 180.0 / M_PI;
   arotX = rotX * 180.0 / M_PI;

      /* rotations for Tsai camera modelling */
   glTranslated ((GLfloat) t_x, (GLfloat) t_y, (GLfloat) t_z);

   glRotated ((GLfloat) arotZ, 0.0, 0.0, 1.0); /* rotation about ?? */
   glRotated ((GLfloat) arotY, 0.0, 1.0, 0.0);
   glRotated ((GLfloat) arotX, 1.0, 0.0, 0.0);

      /* compute proper centering for viewport */
   cXWindow = cXImage*widthWindow/widthImage;
   cYWindow = cYImage*heightWindow/heightImage;
   xWindowD = cXWindow - widthWindow/2.0 ; /* round */
      /* y is flipped so take negative */
   yWindowD = - ( cYWindow - heightWindow/2.0 ) ; /* round */
   xWindow = (xWindowD< 0.0) ? (int)(xWindowD - 0.5):
            (int)(xWindowD + 0.5); /* round */
   yWindow = (yWindowD < 0.0) ? (int)(yWindowD - 0.5):
            (int)(yWindowD + 0.5); /* round */


/*   printf("cXImage %f cYImage %f cXWindow %f cYWindow %f xWindow %d \
 yWindow %d xWindowD %f yWindowD %f\n",
      cXImage, cYImage, cXWindow,cYWindow, xWindow, yWindow,
      xWindowD,yWindowD);
   printf("widthWindow %f heightWindow %f\n",widthWindow, heightWindow);
*/
///   glViewport (0, 0, (GLsizei) widthWindow, (GLsizei) heightWindow);
}

void print(std::string title, GLdouble matrix[16]) {
  std::cout << title;
  for(uint i=0;i<16;i++) {
    if (i%4==0) {std::cout << std::endl; }
    std::cout << matrix[i] << "  ";
  }
  std::cout << std::endl;
}

void Canvas::ViewOptions::addShowOption(Option &opt)
{
  show[opt.name()] = opt.val();
}

bool Canvas::ViewOptions::setShowOption(std::string name, bool value)
{
  return (show[name] = value);

}  

bool Canvas::ViewOptions::invertShowOption(std::string name)
{
  return (show[name] = !show[name]);
}


bool Canvas::ViewOptions::showOption(Option &opt)
{
  return show[opt.name()];
}


Canvas::ViewOptions::ViewOptions(Canvas *canvas) :
  selected_models(),
  last_selection( NULL )
{
  addShowOption(canvas->showBBoxes);
  addShowOption(canvas->showBlocks);
  addShowOption(canvas->showBlur);
  addShowOption(canvas->showClock);
  addShowOption(canvas->showData);
  addShowOption(canvas->showFlags);
  addShowOption(canvas->showFollow);
  addShowOption(canvas->showFootprints);
  addShowOption(canvas->showGrid);
  addShowOption(canvas->showOccupancy);
  addShowOption(canvas->showScreenshots);
  addShowOption(canvas->showStatus);
  addShowOption(canvas->showTrailArrows);
  addShowOption(canvas->showTrailRise);
  addShowOption(canvas->showTrails);
  addShowOption(canvas->showVoxels);
  addShowOption(canvas->pCamOn);
  addShowOption(canvas->showVideo);
  addShowOption(canvas->visualizeAll);
}

void Canvas::ViewOptions::LoadPosition(Worldfile* wf, int sec)
{
  char name[] = "position";
  view_position[0] = wf->ReadTupleFloat( sec, name, 0, 0 )/100.0;
  view_position[1] = wf->ReadTupleFloat( sec, name, 1, 0 )/100.0;
  view_position[2] = wf->ReadTupleFloat( sec, name, 2, 100 )/100.0;
  view_position[3] = wf->ReadTupleFloat( sec, name, 3, 100 )/100.0;
}

void Canvas::ViewOptions::SetViewportParams(GLint w, GLint h)
{
  viewport[0] = view_position[0]*w;
  viewport[1] = view_position[1]*h;
  viewport[2] = view_position[2]*w;
  viewport[3] = view_position[3]*h;
//  printf("viewport: %d %d %d %d\n", viewport[0],viewport[1],viewport[2],viewport[3]);
}

void Canvas::ViewOptions::SetGlViewport()
{
  glViewport(viewport[0],viewport[1],viewport[2],viewport[3]);
};



void Canvas::TimerCallback( Canvas* c )
{
  if( c->world->dirty )
    {
      //puts( "timer redraw" );
      c->redraw();
      c->world->dirty = false;
    }
  
  Fl::repeat_timeout( c->interval/1000.0,
							 (Fl_Timeout_Handler)Canvas::TimerCallback, 
							 c);
}

Canvas::Canvas( WorldGui* world, 
					 int x, int y, 
					 int width, int height) :
  Fl_Gl_Window( x, y, width, height ),
  colorstack(),  
  models_sorted(),
//  current_camera( NULL ),
//  camera(),
//  perspective_camera(),
  dirty_buffer( false ),  
  wf( NULL ),
  startx( -1 ),
  starty( -1 ),
  texture(-1),
  videoname( "" ),
  interval( 40 ), // msec between redraws
  // initialize Option objects
  //  showBlinken( "Blinkenlights", "show_blinkenlights", "", true, world ), 
  showBBoxes( "Debug/Bounding boxes", "show_boundingboxes", "^b", false, world ),
  showBlocks( "Blocks", "show_blocks", "b", true, world ),
  showBlur( "Trails/Blur", "show_trailblur", "^d", false, world ),
  showClock( "Clock", "show_clock", "c", true, world ),
  showData( "Data", "show_data", "d", false, world ),
  showFlags( "Flags", "show_flags", "l",  true, world ),
  showFollow( "Follow", "show_follow", "f", false, world ),
  showFootprints( "Footprints", "show_footprints", "o", false, world ),
  showGrid( "Grid", "show_grid", "g", true, world ),
  showOccupancy( "Debug/Occupancy", "show_occupancy", "^o", false, world ),
  showScreenshots( "Save screenshots", "screenshots", "^e", false, world ),
  showStatus( "Status", "show_status", "s", true, world ),
  showTrailArrows( "Trails/Rising Arrows", "show_trailarrows", "^a", false, world ),
  showTrailRise( "Trails/Rising blocks", "show_trailrise", "^r", false, world ),
  showTrails( "Trails/Fast", "show_trailfast", "^f", false, world ),
  showVoxels( "Debug/Voxels", "show_voxels", "^v", false, world ),
  pCamOn( "Perspective camera", "pcam_on", "r", false, world ),
  showVideo( "Video", "show_video", "v", false, world ),
  visualizeAll( "Selected only", "vis_all", "v", false, world ),
  // and the rest 
  graphics( true ),
  world( world ),
  frames_rendered_count( 0 ),
  screenshot_frame_skip( 1 ),
  selected_view (0)
{
  end();  
  //show(); // must do this so that the GL context is created before configuring GL
  // but that line causes a segfault in Linux/X11! TODO: test in OS X
  PerspectiveCamera pc;
  OrthoCamera oc;
  pc.setPose( 0.0, -4.0, 3.0 );

  perspective_camera_vector.clear();
  camera_vector.clear();
  current_camera_vector.clear();
  camera_vector.push_back(oc);
  perspective_camera_vector.push_back(pc);
  current_camera_vector.push_back(&camera_vector[0]);

  viewOptions.clear();
  ViewOptions vo(this);
  vo.view_position[0] = 0;
  vo.view_position[1] = 0;
  vo.view_position[2] = 1.0;
  vo.view_position[3] = 1.0;
  viewOptions.push_back(vo);

  
  setDirtyBuffer();

  
  // enable accumulation buffer
  //mode( mode() | FL_ACCUM );
  //assert( can_do( FL_ACCUM ) );
}

void Canvas::InitGl()
{
  valid(1);
  FixViewport(w(), h());
  //  LoadGLTextures("rie.bmp",&texture);
  
  // set gl state that won't change every redraw
  glClearColor ( 0.7, 0.7, 0.8, 1.0);
  glDisable( GL_LIGHTING );
  glEnable( GL_DEPTH_TEST );
  glDepthFunc( GL_LESS );
  glCullFace( GL_BACK );
  glEnable (GL_CULL_FACE);
  glEnable( GL_BLEND );
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
  glEnable( GL_LINE_SMOOTH );
  glHint( GL_LINE_SMOOTH_HINT, GL_FASTEST );
  glDepthMask( GL_TRUE );
  glEnable( GL_TEXTURE_2D );
  glEnableClientState( GL_VERTEX_ARRAY );
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  
  // install a font
  gl_font( FL_HELVETICA, 12 );  

  blur = false;
printf("videoname: %s\n",videoname.c_str());
/*
    if ( !videoname.empty() ) {
        texture = TextureManager::getInstance().loadTexture( videoname.c_str() );
    }
*/
  
  // load textures
  std::string fullpath = FileManager::findFile( "assets/stall.png" );
  if ( fullpath == "" ) 
    {
      PRINT_DEBUG( "Unable to load stall texture.\n" );
    }
  
  GLuint stall_id = TextureManager::getInstance().loadTexture( fullpath.c_str() );
  TextureManager::getInstance()._stall_texture_id = stall_id;

  fullpath = FileManager::findFile( "assets/death.png" );
  if ( fullpath == "" )
    {
      PRINT_DEBUG( "Unable to load mine texture.\n" );
    }

  GLuint mine_id = TextureManager::getInstance().loadTexture( fullpath.c_str() );
  TextureManager::getInstance()._mine_texture_id = mine_id;




  
  fullpath = FileManager::findFile( "assets/mainspower.png" );
  if ( fullpath == "" ) 
    {
      PRINT_DEBUG( "Unable to load mains texture.\n" );
    }
  
  GLuint mains_id = TextureManager::getInstance().loadTexture( fullpath.c_str() );
  TextureManager::getInstance()._mains_texture_id = mains_id;
  
  //   // generate a small glow texture
  //   GLubyte* pixels = new GLubyte[ 4 * 128 * 128 ];

  //   for( int x=0; x<128; x++ )
  //  	 for( int y=0; y<128; y++ )
  //  		{		  
  //  		  GLubyte* p = &pixels[ 4 * (128*y + x)];
  //  		  p[0] = (GLubyte)255; // red
  //  		  p[1] = (GLubyte)0; // green
  //  		  p[2] = (GLubyte)0; // blue
  //  		  p[3] = (GLubyte)128; // alpha
  //  		}


  //   glGenTextures(1, &glowTex );
  //   glBindTexture( GL_TEXTURE_2D, glowTex );
  
  //   glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 128, 128, 0, 
  // 					 GL_RGBA, GL_UNSIGNED_BYTE, pixels );

  //   delete[] pixels;

  // draw a check into a bitmap, then load that into a texture
  int i, j;
  for (i = 0; i < checkImageHeight; i++) 
    for (j = 0; j < checkImageWidth; j++) 
      {			
		  int even = (i+j)%2;
		  checkImage[i][j][0] = (GLubyte) 255 - 10*even;
		  checkImage[i][j][1] = (GLubyte) 255 - 10*even;
		  checkImage[i][j][2] = (GLubyte) 255;// - 5*even;
		  checkImage[i][j][3] = 255;
      }


  glGenTextures(1, &checkTex );
  glBindTexture(GL_TEXTURE_2D, checkTex);
  
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, checkImageWidth, checkImageHeight, 
					0, GL_RGBA, GL_UNSIGNED_BYTE, checkImage);
  
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    
  init_done = true; 
}


Canvas::~Canvas()
{ 
  // nothing to do
}

void Canvas::changeView(uint view) {
  showData.set(this, selected_view);
  showBlocks.set(this, selected_view);
  showBBoxes.set(this, selected_view);
  showBlur.set(this, selected_view);
  showClock.set(this, selected_view);
  showFlags.set(this, selected_view);
  showFollow.set(this, selected_view);
  showFootprints.set(this, selected_view);
  showGrid.set(this, selected_view);
  showOccupancy.set(this, selected_view);
  showTrailArrows.set(this, selected_view);
  showTrailRise.set(this, selected_view);
  showTrails.set(this, selected_view);
  showVoxels.set(this, selected_view);
  showScreenshots.set(this, selected_view);
  pCamOn.set(this, selected_view);
  showVideo.set(this, selected_view);

  FOR_EACH(it, world->option_table) {
    (*it)->set(this,selected_view);
//    printf("set %s %d\n", (*it)->name().c_str(), (*it)->val());
      }

  
}

Model* Canvas::getModel( int x, int y )
{
  // render all models in a unique color
  bool showVideo = viewOptions[selected_view].showOption(this->showVideo);

  make_current(); // make sure the GL context is current
  glClearColor( 1,1,1,1 ); // white
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  viewOptions[selected_view].SetGlViewport();
  current_camera_vector[selected_view]->SetProjection();
  current_camera_vector[selected_view]->Draw();

if (showVideo) {

setTsaiCameraView(
-2.297800,
0.277710,
-3.129900,
0.923640,
-0.434690,
3.788700,
7.429700,
1.018600,
2304.000000,
3072.000000,
viewOptions[selected_view].viewport[3],
viewOptions[selected_view].viewport[2],
0.002337,
0.002309,
1562.611100,
929.906800);
}

  
  
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  glDisable(GL_DITHER);
  glDisable(GL_BLEND); // turns off alpha blending, so we read back
  // exactly what we write to a pixel

  // render all top-level, draggable models in a color that is their
  // id 
  FOR_EACH( it, world->World::children )
    {
      Model* mod = (*it);
		
      if( mod->gui.move )
		  {
			 uint8_t rByte, gByte, bByte, aByte;
			 uint32_t modelId = mod->id;
			 rByte = modelId;
			 gByte = modelId >> 8;
			 bByte = modelId >> 16;
			 aByte = modelId >> 24;

//			 printf("mod->Id(): 0x%X, rByte: 0x%X, gByte: 0x%X, bByte: 0x%X, aByte: 0x%X\n", modelId, rByte, gByte, bByte, aByte);
			
			 glColor4ub( rByte, gByte, bByte, aByte );
			 mod->DrawPicker();
		  }
    }
	
  // read the color of the pixel in the back buffer under the mouse
  // pointer
  GLint viewport[4];
  glGetIntegerv(GL_VIEWPORT,viewport);

  uint8_t rByte, gByte, bByte, aByte;
  uint32_t modelId;
	
  glReadPixels( x,h()-y,1,1,
					 GL_RED,GL_UNSIGNED_BYTE,(void*)&rByte );
  glReadPixels( x,h()-y,1,1,
					 GL_GREEN,GL_UNSIGNED_BYTE,(void*)&gByte );
  glReadPixels( x,h()-y,1,1,
					 GL_BLUE,GL_UNSIGNED_BYTE,(void*)&bByte );
  glReadPixels( x,h()-y,1,1,
					 GL_ALPHA,GL_UNSIGNED_BYTE,(void*)&aByte );
	
  modelId = rByte;
  modelId |= gByte << 8;
  modelId |= bByte << 16;
  //modelId |= aByte << 24;
	
//  	printf("Clicked rByte: 0x%X, gByte: 0x%X, bByte: 0x%X, aByte: 0x%X\n", rByte, gByte, bByte, aByte);
//  	printf("-->model Id = 0x%X at [%d %d]\n", modelId,x,y);
	
  Model* mod = Model::LookupId( modelId );

  //printf("%p %s %d %x\n", mod, mod ? mod->Token() : "(none)", id, id );

  // put things back the way we found them
  glEnable(GL_DITHER);
  glEnable(GL_BLEND);
  glClearColor ( 0.7, 0.7, 0.8, 1.0);

  
  // useful for debugging the picker
  //Screenshot();

/*glMatrixMode(GL_PROJECTION);
glPopMatrix();
glMatrixMode(GL_MODELVIEW);
glPopMatrix();*/
  
  return mod;
}

bool Canvas::selected( Model* mod ) 
{
  return( std::find( viewOptions[selected_view].selected_models.begin(),
                     viewOptions[selected_view].selected_models.end(), mod ) != viewOptions[selected_view].selected_models.end() );
}

void Canvas::select( Model* mod ) {
  if( mod )
    {
      viewOptions[selected_view].last_selection = mod;
      viewOptions[selected_view].selected_models.push_front( mod );
		
      //		mod->Disable();
      redraw();
    }
}

void Canvas::unSelect( Model* mod ) 
{
  if( mod )
    {
		EraseAll( mod, viewOptions[selected_view].selected_models );
      redraw();
    }
}

void Canvas::unSelectAll() 
{ 
  viewOptions[selected_view].selected_models.clear();
}

int Canvas::getView(int mx, int my)
{
  int x, y, w, h;
  int result = -1;

  for(uint i=0; i<viewOptions.size();i++) {
    x = viewOptions[i].viewport[0];
    y = viewOptions[i].viewport[1];
    w = viewOptions[i].viewport[2];
    h = viewOptions[i].viewport[3];
    if (x < mx && mx < x+w && y < my && my < y+h) {
      result = i;
      break;
    }
  }
  return result;
}
  
// convert from 2d window pixel to 3d world coordinates
void Canvas::CanvasToWorld( int px, int py, 
									 double *wx, double *wy, double* wz, uint view )
{
  int x = viewOptions[view].viewport[0];
  int y = viewOptions[view].viewport[1];
  int w = viewOptions[view].viewport[2];
  int hh = viewOptions[view].viewport[3];
  int pyo = h()-py-y;
  int pxo = px-x;
  bool showVideo = viewOptions[view].showOption(this->showVideo);


//printf("%d %d\n",px,py);
  
  if( pxo <= 0 )
    pxo = 1;
  else if( pxo >= w )
    pxo = w - 1;
  if( pyo <= 0 )
    pyo = 1;
  else if( pyo >= hh )
    pyo = hh - 1;
	
  //redraw the screen only if the camera model isn't active.
  //TODO new selection technique will simply use drawfloor to result in z = 0 always and prevent strange behaviours near walls
  //TODO refactor, so glReadPixels reads (then caches) the whole screen only when the camera changes.
  if( true || dirtyBuffer() ) {
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    viewOptions[selected_view].SetGlViewport();
    current_camera_vector[view]->SetProjection();
    current_camera_vector[view]->Draw();


if (showVideo) {
setTsaiCameraView(
-2.297800,
0.277710,
-3.129900,
0.923640,
-0.434690,
3.788700,
7.429700,
1.018600,
2304.000000,
3072.000000,
hh,
w,
0.002337,
0.002309,
1562.611100,
929.906800);
}

    
    DrawFloor(); //call this rather than renderFrame for speed - this won't give correct z values
    dirty_buffer = false;
  }
	
  int viewport[4];
  glGetIntegerv(GL_VIEWPORT, viewport);

  GLdouble modelview[16];
  glGetDoublev(GL_MODELVIEW_MATRIX, modelview);

  GLdouble projection[16];	
  glGetDoublev(GL_PROJECTION_MATRIX, projection);

  GLfloat pz;
  glReadPixels( pxo, pyo, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &pz );
  gluUnProject( pxo, pyo, pz, modelview, projection, viewport, wx,wy,wz );
}

int Canvas::handle(int event) 
{
  bool pCamOn = viewOptions[selected_view].showOption(this->pCamOn);

  switch(event)
    {
    case FL_MOUSEWHEEL:
      if(  pCamOn == true ) {
		  perspective_camera_vector[selected_view].scroll( Fl::event_dy() / 10.0 );
      }
      else {
        int ww = viewOptions[selected_view].viewport[2];
        int hh = viewOptions[selected_view].viewport[3];
		  camera_vector[selected_view].scale( Fl::event_dy(),  Fl::event_x(), ww, Fl::event_y(), hh );
      }
      invalidate();
      redraw();
      return 1;
		
    case FL_MOVE: // moused moved while no button was pressed
      if( Fl::event_state( FL_META ) )
		  {
			 puts( "TODO: HANDLE HISTORY" );
			 //world->paused = ! world->paused;
			 return 1;
		  }
		
      if ( startx >=0 ) 
		  {
			 // mouse pointing to valid value
			 
			 if( Fl::event_state( FL_CTRL ) )
				{
				  int dx = Fl::event_x() - startx;
				  int dy = Fl::event_y() - starty;
				  
				  if( pCamOn == true ) {
					 perspective_camera_vector[selected_view].addYaw( -dx );
					 perspective_camera_vector[selected_view].addPitch( -dy );
				  } 
				  else {
					 camera_vector[selected_view].addPitch( - 0.5 * static_cast<double>( dy ) );
					 camera_vector[selected_view].addYaw( - 0.5 * static_cast<double>( dx ) );
				  }
				  invalidate();
				  redraw();
				}
			 else if( Fl::event_state( FL_ALT ) )
				{   
				  int dx = Fl::event_x() - startx;
				  int dy = Fl::event_y() - starty;
				  
				  if( pCamOn == true ) {
					 perspective_camera_vector[selected_view].move( -dx, dy, 0.0 );
				  } 
				  else {
					 camera_vector[selected_view].move( -dx, dy );
				  }
				  invalidate();
				}
		  }
      startx = Fl::event_x();
      starty = Fl::event_y();
      return 1;
    case FL_PUSH: // button pressed
      {
        if (Fl::event_button() == 1)
        {
          int v = getView(startx,h()-starty);
          if (v != -1 && (int) selected_view != v) {
            selected_view = v;
            changeView(v);
          }
        }

        {
 			 Model* mod = getModel( startx, starty );
			 startx = Fl::event_x();
			 starty = Fl::event_y();
			 viewOptions[selected_view].selectedModel = false;
			 switch( Fl::event_button() )
				{
				case 1:
				  clicked_empty_space = ( mod == NULL );
				  empty_space_startx = startx;
				  empty_space_starty = starty;
              
				  if( mod ) { 
					 // clicked a model
					 if ( Fl::event_state( FL_SHIFT ) ) {
						// holding shift, toggle selection
						if ( selected( mod ) ) 
						  unSelect( mod );
						else {
						  select( mod );
						  viewOptions[selected_view].selectedModel = true; // selected a model
						}
					 }
					 else {
						if ( !selected( mod ) ) {
						  // clicked on an unselected model while
						  //  not holding shift, this is the new
						  //  selection
						  unSelectAll();
						  select( mod );
						}
						viewOptions[selected_view].selectedModel = true; // selected a model
					 }
				  }
				  
				  redraw(); // probably required	
				  return 1;
				case 3:
				  {
					 // leave selections alone
					 // rotating handled within FL_DRAG
					 return 1;
				  }
				default:
				  return 0;
				}    
		  }
      }
	  
    case FL_DRAG: // mouse moved while button was pressed
      {
		  int dx = Fl::event_x() - startx;
		  int dy = Fl::event_y() - starty;

		  if ( Fl::event_state( FL_BUTTON1 ) && Fl::event_state( FL_CTRL ) == false ) {
			 // Left mouse button drag
			 if ( viewOptions[selected_view].selectedModel ) {
				// started dragging on a selected model
				
				double sx,sy,sz;
				CanvasToWorld( startx, starty,
									&sx, &sy, &sz, selected_view );
				double x,y,z;
				CanvasToWorld( Fl::event_x(), Fl::event_y(),
									&x, &y, &z, selected_view );
				// move all selected models to the mouse pointer
				FOR_EACH( it, viewOptions[selected_view].selected_models )
				  {
					 Model* mod = *it;
					 mod->AddToPose( x-sx, y-sy, 0, 0 );
				  }
			 }
			 else {
				// started dragging on empty space or an
				//  unselected model, move the canvas
				if( pCamOn == true ) {
				  perspective_camera_vector[selected_view].move( -dx, dy, 0.0 );
				} 
				else {
				  camera_vector[selected_view].move( -dx, dy );
				}
				invalidate(); // so the projection gets updated
			 }
		  }
		  else if ( Fl::event_state( FL_BUTTON3 ) || ( Fl::event_state( FL_BUTTON1 ) &&  Fl::event_state( FL_CTRL )  ) ) {
			 // rotate all selected models
			 FOR_EACH( it, viewOptions[selected_view].selected_models )
				{
				  Model* mod = *it;
				  mod->AddToPose( 0,0,0, 0.05*(dx+dy) );
				}
		  }
		
		  startx = Fl::event_x();
		  starty = Fl::event_y();

		  redraw();
	  return 1;
      } // end case FL_DRAG

    case FL_RELEASE:   // mouse button released
      if( empty_space_startx == Fl::event_x() && empty_space_starty == Fl::event_y() && clicked_empty_space == true ) {
		  // clicked on empty space, unselect all
		  unSelectAll();
		  redraw();
      }
      return 1;

    case FL_FOCUS:
    case FL_UNFOCUS:
      //.... Return 1 if you want keyboard events, 0 otherwise
    return 1;

    case FL_KEYBOARD:
      switch( Fl::event_key() )
		  {
		  case FL_Left:
			 if( pCamOn == false ) { camera_vector[selected_view].move( -10, 0 ); } 
			 else { perspective_camera_vector[selected_view].strafe( -0.5 ); } break;
		  case FL_Right: 
			 if( pCamOn == false ) {camera_vector[selected_view].move( 10, 0 ); } 
			 else { perspective_camera_vector[selected_view].strafe( 0.5 ); } break;
		  case FL_Down:  
			 if( pCamOn == false ) {camera_vector[selected_view].move( 0, -10 ); } 
			 else { perspective_camera_vector[selected_view].forward( -0.5 ); } break;
		  case FL_Up:  
			 if( pCamOn == false ) {camera_vector[selected_view].move( 0, 10 ); } 
			 else { perspective_camera_vector[selected_view].forward( 0.5 ); } break;
        case FL_Tab:
          selected_view = (selected_view+1) % viewOptions.size();
          printf("selected view: %d\n",selected_view);
          changeView(selected_view);
          redraw(); // we probably set a display config - so need this
          break;
		  default:
			 redraw(); // we probably set a display config - so need this
			 return 0; // keypress unhandled
		  }
		
      invalidate(); // update projection
      return 1;
			
      //	case FL_SHORTCUT:
      //		//... shortcut, key is in Fl::event_key(), ascii in Fl::event_text()
      //		//... Return 1 if you understand/use the shortcut event, 0 otherwise...
      //		return 1;
    default:
      // pass other events to the base class...
      //printf( "EVENT %d\n", event );
      return Fl_Gl_Window::handle(event);
			
    } // end switch( event )
  return 0;
}

void Canvas::FixViewport(int W,int H) 
{
  glLoadIdentity();
  glViewport(0,0,W,H);
}

void Canvas::AddModel( Model*  mod  )
{
  models_sorted.push_back( mod );
	redraw();
}

void Canvas::RemoveModel( Model*  mod  )
{
  printf( "removing model %s from canvas list\n", mod->Token() );
  EraseAll( mod, models_sorted );
}

void Canvas::DrawGlobalGrid(uint view)
{

  stg_bounds3d_t bounds = world->GetExtent();

  /*   printf( "bounds [%.2f %.2f] [%.2f %.2f] [%.2f %.2f]\n",
       bounds.x.min, bounds.x.max,
       bounds.y.min, bounds.y.max,
       bounds.z.min, bounds.z.max );
			 
  */
  
  /* simple scaling of axis labels - could be better */
  int skip = (int)( 50 / camera_vector[view].scale());
  if( skip < 1 ) skip = 1;
  if( skip > 2 && skip % 2 ) skip += 1;

  //printf( "scale %.4f\n", camera.scale() );

  char str[64];	
  PushColor( 0.2, 0.2, 0.2, 1.0 ); // pale gray

  for( double i=0; i < bounds.x.max; i+=skip)
    {
      snprintf( str, 16, "%d", (int)i );
      Gl::draw_string(  i, 0, 0.00, str );
    }

  for( double i=0; i >= bounds.x.min; i-=skip)
    {
      snprintf( str, 16, "%d", (int)i );
      Gl::draw_string(  i, 0, 0.00, str );
    }

  
  for( double i=0; i < bounds.y.max; i+=skip)
    {
      snprintf( str, 16, "%d", (int)i );
      Gl::draw_string(  0, i, 0.00, str );
    }

  for( double i=0; i >= bounds.y.min; i-=skip)
    {
      snprintf( str, 16, "%d", (int)i );
      Gl::draw_string(  0, i, 0.00, str );
    }


  PopColor();
  
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(2.0, 2.0);
  glDisable(GL_BLEND);

  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, checkTex );
  glColor3f( 1.0, 1.0, 1.0 );

  glBegin(GL_QUADS);
  glTexCoord2f( bounds.x.min/2.0, bounds.y.min/2.0 ); 
  glVertex2f( bounds.x.min, bounds.y.min );
  glTexCoord2f( bounds.x.max/2.0, bounds.y.min/2.0); 
  glVertex2f(  bounds.x.max, bounds.y.min );
  glTexCoord2f( bounds.x.max/2.0, bounds.y.max/2.0 ); 
  glVertex2f(  bounds.x.max, bounds.y.max );
  glTexCoord2f( bounds.x.min/2.0, bounds.y.max/2.0 ); 
  glVertex2f( bounds.x.min, bounds.y.max );
  glEnd();

  glDisable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);
  
  glDisable(GL_POLYGON_OFFSET_FILL );
}

//draw the floor without any grid ( for robot's perspective camera model )
void Canvas::DrawFloor()
{
  stg_bounds3d_t bounds = world->GetExtent();
	
  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(2.0, 2.0);
	
  glColor4f( 1.0, 1.0, 1.0, 1.0 );

  glBegin(GL_QUADS);
  glVertex2f( bounds.x.min, bounds.y.min );
  glVertex2f( bounds.x.max, bounds.y.min );
  glVertex2f( bounds.x.max, bounds.y.max );
  glVertex2f( bounds.x.min, bounds.y.max );
  glEnd();
	
  glEnd();
}

void Canvas::DrawBlocks() 
{
  FOR_EACH( it, models_sorted )
    (*it)->DrawBlocksTree();
}

void Canvas::DrawBoundingBoxes() 
{
  glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
  glLineWidth( 2.0 );
  glPointSize( 5.0 );
  glDisable (GL_CULL_FACE);
  
  world->DrawBoundingBoxTree();
  
  glEnable (GL_CULL_FACE);
  glLineWidth( 1.0 );
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
}

void Canvas::resetCamera(uint view)
{
  float max_x = 0, max_y = 0, min_x = 0, min_y = 0;
  int w = viewOptions[view].viewport[2];
  int h = viewOptions[view].viewport[3];
	
  //TODO take orrientation ( `a' ) and geom.pose offset into consideration
  FOR_EACH( it, world->World::children )
    {	 
      Model* ptr = (*it);
      Pose pose = ptr->GetPose();
      Geom geom = ptr->GetGeom();
		
      float tmp_min_x = pose.x - geom.size.x / 2.0;
      float tmp_max_x = pose.x + geom.size.x / 2.0;
      float tmp_min_y = pose.y - geom.size.y / 2.0;
      float tmp_max_y = pose.y + geom.size.y / 2.0;
		
      if( tmp_min_x < min_x ) min_x = tmp_min_x;
      if( tmp_max_x > max_x ) max_x = tmp_max_x;
      if( tmp_min_y < min_y ) min_y = tmp_min_y;
      if( tmp_max_y > max_y ) max_y = tmp_max_y;
    }
	
  //do a complete reset
  float x = ( min_x + max_x ) / 2.0;
  float y = ( min_y + max_y ) / 2.0;
  camera_vector[view].setPose( x, y );
  float scale_x = w / (max_x - min_x) * 0.9;
  float scale_y = h / (max_y - min_y) * 0.9;
  camera_vector[view].setScale( scale_x < scale_y ? scale_x : scale_y );
	
  //TODO reset perspective cam
}


class DistFuncObj
{
  stg_meters_t x, y;
  DistFuncObj( stg_meters_t x, stg_meters_t y ) 
    : x(x), y(y) {}
  
  bool operator()(const Model* a, const Model* b ) const
  { 
    Pose a_pose = a->GetGlobalPose();
    Pose b_pose = b->GetGlobalPose();
	 
    stg_meters_t a_dist = hypot( y - a_pose.y,
											x - a_pose.x );
	 
    stg_meters_t b_dist = hypot( y - b_pose.y,
											x - b_pose.x );
	 
    return (  a_dist < b_dist );
  }
};


void Canvas::renderFrame(uint view)
{
  bool showData = viewOptions[view].showOption(this->showData);
  bool showBlocks = viewOptions[view].showOption(this->showBlocks);
  bool showFlags = viewOptions[view].showOption(this->showFlags);
  bool showClock = viewOptions[view].showOption(this->showClock);
  bool showFollow = viewOptions[view].showOption(this->showFollow);
  bool showFootprints = viewOptions[view].showOption(this->showFootprints);
  bool showGrid = viewOptions[view].showOption(this->showGrid);
  bool showStatus = viewOptions[view].showOption(this->showStatus);
  bool showOccupancy = viewOptions[view].showOption(this->showOccupancy);
  bool showTrailArrows = viewOptions[view].showOption(this->showTrailArrows);
  bool showTrails = viewOptions[view].showOption(this->showTrails);
  bool showTrailRise = viewOptions[view].showOption(this->showTrailRise);
  bool showBBoxes = viewOptions[view].showOption(this->showBBoxes);
  bool showVoxels = viewOptions[view].showOption(this->showVoxels);
  bool showScreenshots = viewOptions[view].showOption(this->showScreenshots);
  bool visualizeAll = viewOptions[view].showOption(this->visualizeAll);
  bool showVideo = viewOptions[view].showOption(this->showVideo);


  //before drawing, order all models based on distance from camera
  float x = current_camera_vector[view]->x();
  float y = current_camera_vector[view]->y();
  float sphi = -dtor( current_camera_vector[view]->yaw() );
	
  //estimate point of camera location - hard to do with orthogonal mode
  x += -sin( sphi ) * 100;
  y += -cos( sphi ) * 100;
	
  //double coords[2];
  //coords[0] = x;
  //coords[1] = y;
  
  // sort the list of models by inverse distance from the camera -
  // probably doesn't change too much between frames so this is
  // usually fast
  // TODO
  //models_sorted = g_list_sort_with_data( models_sorted, (GCompareDataFunc)compare_distance, coords );
  
  // TODO: understand why this doesn't work and fix it - cosmetic but important!
  //std::sort( models_sorted.begin(), models_sorted.end(), DistFuncObj(x,y) );

  glEnable( GL_DEPTH_TEST );

///  if( ! showTrails )
///    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  if (showVideo) {
//    displayTexture(texture);
    displayVideo();

  }
//  drawpath();
  
  if( showOccupancy )
	 ((WorldGui*)world)->DrawOccupancy();
  
  if( showVoxels )
	 ((WorldGui*)world)->DrawVoxels();
  
  
  if( ! world->rt_cells.empty() )
    {
      glPushMatrix();	  		
      GLfloat scale = 1.0/world->Resolution();
      glScalef( scale, scale, 1.0 ); // XX TODO - this seems slightly
			
      world->PushColor( Color( 0,0,1,0.5) );
			
      glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
			
      glPointSize( 2 );
      glBegin( GL_POINTS );
			
      for( unsigned int i=0;
			  i < world->rt_cells.size();
			  i++ )
		  {
			 char str[128];
			 snprintf( str, 128, "(%d,%d)", 
						  world->rt_cells[i].x, 
						  world->rt_cells[i].y );
					
			 Gl::draw_string(  world->rt_cells[i].x+1, 
									 world->rt_cells[i].y+1, 0.1, str );
					
			 //printf( "x: %d y: %d\n", world->rt_regions[i].x, world->rt_regions[i].y );
			 //glRectf( world->rt_cells[i].x+0.3, world->rt_cells[i].y+0.3,
			 //	 world->rt_cells[i].x+0.7, world->rt_cells[i].y+0.7 );
					
			 glVertex2f( world->rt_cells[i].x, world->rt_cells[i].y );
					
		  }
			
      glEnd();
			
#if 1
      world->PushColor( Color( 0,1,0,0.2) );
      glBegin( GL_LINE_STRIP );
      for( unsigned int i=0;
			  i < world->rt_cells.size();
			  i++ )
		  {			 
			 glVertex2f( world->rt_cells[i].x+0.5, world->rt_cells[i].y+0.5 );
		  }
      glEnd();
      world->PopColor();
#endif
			
      glPopMatrix();
      world->PopColor();
    }
	
  if( ! world->rt_candidate_cells.empty() )
    {
      glPushMatrix();	  		
      GLfloat scale = 1.0/world->Resolution();
      glScalef( scale, scale, 1.0 ); // XX TODO - this seems slightly
			
      world->PushColor( Color( 1,0,0,0.5) );
			
      glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
			
      for( unsigned int i=0;
			  i < world->rt_candidate_cells.size();
			  i++ )
		  {
			 // 			 char str[128];
			 // 			 snprintf( str, 128, "(%d,%d)", 
			 // 						  world->rt_candidate_cells[i].x, 
			 // 						  world->rt_candidate_cells[i].y );
					
			 // 			 Gl::draw_string(  world->rt_candidate_cells[i].x+1, 
			 // 									 world->rt_candidate_cells[i].y+1, 0.1, str );
					
			 //printf( "x: %d y: %d\n", world->rt_regions[i].x, world->rt_regions[i].y );
			 glRectf( world->rt_candidate_cells[i].x, world->rt_candidate_cells[i].y,
						 world->rt_candidate_cells[i].x+1, world->rt_candidate_cells[i].y+1 );
		  }
			
      world->PushColor( Color( 0,1,0,0.2) );
      glBegin( GL_LINE_STRIP );
      for( unsigned int i=0;
			  i < world->rt_candidate_cells.size();
			  i++ )
		  {			 
			 glVertex2f( world->rt_candidate_cells[i].x+0.5, world->rt_candidate_cells[i].y+0.5 );
		  }
      glEnd();
      world->PopColor();
			
      glPopMatrix();
      world->PopColor();
			
      //world->rt_cells.clear();
    }

//TODO: if !showVideo
  if ( !showVideo ) {
    if( showGrid )
      DrawGlobalGrid(view);
    else
      DrawFloor();
  }
  
  if( showFootprints )
    {
      glDisable( GL_DEPTH_TEST ); // using alpha blending		
		
      FOR_EACH( it, models_sorted )
		  (*it)->DrawTrailFootprint();
		
      glEnable( GL_DEPTH_TEST );
    }
  
  if( showFlags ) 
    FOR_EACH( it, models_sorted )
      (*it)->DrawFlagList();

  if( showTrailArrows )
    FOR_EACH( it, models_sorted )
      (*it)->DrawTrailArrows();  
  
  if( showTrailRise )
    FOR_EACH( it, models_sorted )
      (*it)->DrawTrailBlocks();  
    
  if( showBlocks )
    DrawBlocks();
	
  if( showBBoxes )
    DrawBoundingBoxes();
	
  // TODO - finish this properly
  //LISTMETHOD( models_sorted, Model*, DrawWaypoints );

  FOR_EACH( it, viewOptions[view].selected_models )
    (*it)->DrawSelected();

  // useful debug - puts a point at the origin of each model
  //for( GList* it = world->World::children; it; it=it->next ) 
  // ((Model*)it->data)->DrawOriginTree();
  
  // draw the model-specific visualizations
  if( showData ) {
    if ( ! visualizeAll ) {
      FOR_EACH( it, world->World::children ) {
		  (*it)->DataVisualizeTree( current_camera_vector[view] );
      }
    }
    else if ( viewOptions[view].selected_models.size() > 0 ) {
      FOR_EACH( it, world->World::children ) {
		  (*it)->DataVisualizeTree( current_camera_vector[view] );
      }
    }
    else if ( viewOptions[view].last_selection ) {
      viewOptions[view].last_selection->DataVisualizeTree( current_camera_vector[view] );
    }
  }
   
  if( showGrid ) 
    FOR_EACH( it, models_sorted )
      (*it)->DrawGrid();
  		  
  if( showStatus ) 
    {
      glPushMatrix();
      //ensure two icons can't be in the exact same plane
      if( camera_vector[view].pitch() == 0 && !viewOptions[view].showOption(pCamOn) )
		  glTranslatef( 0, 0, 0.1 );
		
      FOR_EACH( it, models_sorted )
		  (*it)->DrawStatusTree( &camera_vector[view] );
		
      glPopMatrix();
    }
  
  if( world->ray_list.size() > 0 )
    {
      glDisable( GL_DEPTH_TEST );
      PushColor( 0,0,0,0.5 );
      FOR_EACH( it, world->ray_list )
		  {
			 float* pts = *it;
			 glBegin( GL_LINES );
			 glVertex2f( pts[0], pts[1] );
			 glVertex2f( pts[2], pts[3] );
			 glEnd();
		  }  
      PopColor();
      glEnable( GL_DEPTH_TEST );
		 
      world->ClearRays();
    } 
	
  if( showClock && view==0 ) // view 0 should be main window
    {		
      glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
		
      //use orthogonal projeciton without any zoom
      glMatrixMode (GL_PROJECTION);
      glPushMatrix(); //save old projection
      glLoadIdentity ();
      glOrtho( 0, w(), 0, h(), -100, 100 );	
      glMatrixMode (GL_MODELVIEW);

      glPushMatrix();
      glLoadIdentity();
      glDisable( GL_DEPTH_TEST );

      std::string clockstr = world->ClockString();
      if( showFollow == true && viewOptions[view].last_selection )
		  clockstr.append( " [FOLLOW MODE]" );
		
      float txtWidth = gl_width( clockstr.c_str());
      if( txtWidth < 200 ) txtWidth = 200;
      int txtHeight = gl_height();
		
      const int margin = 5;
      int width, height;		
      width = txtWidth + 2 * margin;
      height = txtHeight + 2 * margin; 
		
      // TIME BOX
      colorstack.Push( 0.8,0.8,1.0 ); // pale blue
      glRectf( 0, 0, width, height );
      colorstack.Push( 0,0,0 ); // black
      Gl::draw_string( margin, margin, 0, clockstr.c_str() );
      colorstack.Pop();
      colorstack.Pop();
		
      // ENERGY BOX
      if( PowerPack::global_capacity > 0 )
		  {
			 colorstack.Push( 0.8,1.0,0.8,0.85 ); // pale green
			 glRectf( 0, height, width, 90 );
			 colorstack.Push( 0,0,0 ); // black
			 Gl::draw_string_multiline( margin, height + margin, width, 50, 
												 world->EnergyString().c_str(), 
												 (Fl_Align)( FL_ALIGN_LEFT | FL_ALIGN_BOTTOM) );	 
			 colorstack.Pop();
			 colorstack.Pop();
		  }
			 
      glEnable( GL_DEPTH_TEST );
      glPopMatrix();

      //restore camera projection
      glMatrixMode (GL_PROJECTION);
      glPopMatrix();
      glMatrixMode (GL_MODELVIEW);
    }

  if( showScreenshots && (frames_rendered_count % screenshot_frame_skip == 0) )
    Screenshot();

  frames_rendered_count++; 
}


void Canvas::EnterScreenCS()
{
  //use orthogonal projection without any zoom
  glMatrixMode (GL_PROJECTION);
  glPushMatrix(); //save old projection
  glLoadIdentity ();
  glOrtho( 0, w(), 0, h(), -100, 100 );	
  glMatrixMode (GL_MODELVIEW);
  
  glPushMatrix();
  glLoadIdentity();
  glDisable( GL_DEPTH_TEST );
}

void Canvas::LeaveScreenCS()
{
  glEnable( GL_DEPTH_TEST );
  glPopMatrix();
  glMatrixMode (GL_PROJECTION);
  glPopMatrix();
  glMatrixMode (GL_MODELVIEW);
}


void Canvas::Screenshot(std::string name)
{

  int width = w();
  int height = h();
  int depth = 4; // RGBA

  // we use RGBA throughout, though we only need RGB, as the 4-byte
  // pixels avoid a nasty word-alignment problem when indexing into
  // the pixel array.

  // might save a bit of time with a static var as the image size rarely changes
  static std::vector<uint8_t> pixels;
  pixels.resize( width * height * depth );
  
  glFlush(); // make sure the drawing is done
  // read the pixels from the screen
  glReadPixels( 0,0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, &pixels[0] );			 
  
  static uint32_t count = 0;		 
  char filename[301];
  if (name.empty()) {
    snprintf( filename, 300, "stage-%05d.png", count++ );
  } else {
    strncpy(filename,name.c_str(),300);
  }

  
  FILE *fp = fopen( filename, "wb" );
  if( fp == NULL ) 
    {
      PRINT_ERR1( "Unable to open %s", filename );
    }
  
  // create PNG data
  png_structp pp = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
  assert(pp);
  png_infop info = png_create_info_struct(pp);
  assert(info);

  // setup the output file
  png_init_io(pp, fp);

  // need to invert the image as GL and PNG disagree on the row order
  png_bytep rowpointers[height];
  for( int i=0; i<height; i++ )
    rowpointers[i] = &pixels[ (height-1-i) * width * depth ];

  png_set_rows( pp, info, rowpointers ); 

  png_set_IHDR( pp, info, 
					 width, height, 8, 
					 PNG_COLOR_TYPE_RGBA, 
					 PNG_INTERLACE_NONE, 
					 PNG_COMPRESSION_TYPE_DEFAULT, 
					 PNG_FILTER_TYPE_DEFAULT);

  png_write_png( pp, info, PNG_TRANSFORM_IDENTITY, NULL );

  // free the PNG data - we reuse the pixel array next call.
  png_destroy_write_struct(&pp, &info);
  
  fclose(fp);
  
  printf( "Saved %s\n", filename );
}

void Canvas::perspectiveCb( Fl_Widget* w, void* p ) 
{
  Canvas* canvas = static_cast<Canvas*>( w );
  Option* opt = static_cast<Option*>( p ); // pCamOn
///  canvas->viewOptions[canvas->selected_view].pCamOn = opt;
  if ( opt ) {
    // Perspective mode is on, change camera
    canvas->current_camera_vector[canvas->selected_view] = &canvas->perspective_camera_vector[canvas->selected_view];
  }
  else {
    canvas->current_camera_vector[canvas->selected_view] = &canvas->camera_vector[canvas->selected_view];
  }
	
  canvas->invalidate();
}

void Canvas::createMenuItems( Fl_Menu_Bar* menu, std::string path )
{
  showData.createMenuItem( menu, path );
  //  visualizeAll.createMenuItem( menu, path );
  showBlocks.createMenuItem( menu, path );
  showFlags.createMenuItem( menu, path );
  showClock.createMenuItem( menu, path );
  showFlags.createMenuItem( menu, path );
  showFollow.createMenuItem( menu, path );
  showFootprints.createMenuItem( menu, path );
  showGrid.createMenuItem( menu, path );
  showStatus.createMenuItem( menu, path );
  pCamOn.createMenuItem( menu, path );
  pCamOn.menuCallback( perspectiveCb, this );
  showOccupancy.createMenuItem( menu, path );
  showTrailArrows.createMenuItem( menu, path );
  showTrails.createMenuItem( menu, path ); 
  showTrailRise.createMenuItem( menu, path );  // broken
  showBBoxes.createMenuItem( menu, path );
  showVoxels.createMenuItem( menu, path );  
  showScreenshots.createMenuItem( menu, path );
  showVideo.createMenuItem( menu, path );
  showVideo.menuCallback( videoCb, this );

}


void Canvas::Load( Worldfile* wf, int sec )
{
  this->wf = wf;

  selected_view = 0;
  PerspectiveCamera pc;
  OrthoCamera oc;
  pc.setPose( 0.0, -4.0, 3.0 );


  bool views_defined = false;
  
  for( int entity = 1; entity < wf->GetEntityCount(); entity++ )
    if( wf->GetEntityParent(entity) == sec && strcmp(wf->GetEntityType(entity),"view")==0 )
    {
      if (!views_defined) {
        views_defined = true;
        perspective_camera_vector.clear();
        camera_vector.clear();
        viewOptions.clear();
        current_camera_vector.clear();
      }
      ViewOptions vo(this);
      camera_vector.push_back(oc);
      perspective_camera_vector.push_back(pc);
      vo.LoadPosition(wf,entity);
      showData.Load( wf, entity );
      showFlags.Load( wf, entity );
      showBlocks.Load( wf, entity );
      showBBoxes.Load( wf, entity );
      showBlur.Load( wf, entity );
      showClock.Load( wf, entity );
      showFollow.Load( wf, entity );
      showFootprints.Load( wf, entity );
      showGrid.Load( wf, entity );
      showOccupancy.Load( wf, entity );
      showTrailArrows.Load( wf, entity );
      showTrailRise.Load( wf, entity );
      showTrails.Load( wf, entity );
      showVoxels.Load( wf, entity );
      showScreenshots.Load( wf,  entity);
      pCamOn.Load( wf, entity );
      showVideo.Load( wf,  entity);
      vo.addShowOption(showData);
      vo.addShowOption(showFlags);
      vo.addShowOption(showBlocks);
      vo.addShowOption(showBBoxes);
      vo.addShowOption(showBlur);
      vo.addShowOption(showClock);
      vo.addShowOption(showFollow);
      vo.addShowOption(showFootprints);
      vo.addShowOption(showGrid);
      vo.addShowOption(showOccupancy);
      vo.addShowOption(showTrailArrows);
      vo.addShowOption(showTrailRise);
      vo.addShowOption(showTrails);
      vo.addShowOption(showVoxels);
      vo.addShowOption(showScreenshots);
      vo.addShowOption(pCamOn);
      vo.addShowOption(showVideo);
//puts("optins:");
      FOR_EACH(it, world->option_table) {
//	printf("set %s %d\n", (*it)->name().c_str(), (*it)->val());
          (*it)->Load(wf,entity);
          vo.addShowOption(**it);
      }
      vo.show["Graphics"] = true; //hack to register custom
      viewOptions.push_back(vo);
      

      std::string model_str = wf->ReadString( entity, "last_selection", "");
      if (!model_str.empty()) {
        Model *mod = world->GetModel( model_str );
        if( mod )
        {
          viewOptions[selected_view].last_selection = mod;
//          viewOptions[selected_view].selected_models.push_front( mod );
        }
      }
      selected_view++;
    }
    selected_view = 0;



for(uint i=0;i<camera_vector.size();i++) {
      current_camera_vector.push_back(&camera_vector[i]);
    }

    videoname = wf->ReadString( sec, "video", "");
printf("reading videoname: %s\n",videoname.c_str());
init_done = false;
    //TODO: load all cameras
    camera_vector[0].Load( wf, sec );
    perspective_camera_vector[0].Load( wf, sec );

    interval = wf->ReadInt(sec, "interval", interval );

    screenshot_frame_skip = wf->ReadInt( sec, "screenshot_skip", screenshot_frame_skip );
    if( screenshot_frame_skip < 1 )
      screenshot_frame_skip = 1; // avoids div-by-zero if poorly set


      if( ! world->paused )
        // // start the timer that causes regular redraws
        Fl::add_timeout( ((double)interval/1000),
                         (Fl_Timeout_Handler)Canvas::TimerCallback,
                         this);
                         invalidate(); // we probably changed something

//FOR_EACH(it,world->models_by_name) {
//  printf("model: %s\n",it->first.c_str());
//}

 if (showVideo) {
  videoCb(this, 0 );
 }
}

void Canvas::Save( Worldfile* wf, int sec )
{
  //TODO: save all cameras & all view options
  
  puts("Window saving not supported.");
/*
  camera_vector[0].Save( wf, sec );
  perspective_camera_vector[0].Save( wf, sec );	
	
  wf->WriteInt(sec, "interval", interval );

  showData.Save( wf, sec );
  showBlocks.Save( wf, sec );
  showBBoxes.Save( wf, sec );
  showBlur.Save( wf, sec );
  showClock.Save( wf, sec );
  showFlags.Save( wf, sec );
  showFollow.Save( wf, sec );
  showFootprints.Save( wf, sec );
  showGrid.Save( wf, sec );
  showOccupancy.Save( wf, sec );
  showTrailArrows.Save( wf, sec );
  showTrailRise.Save( wf, sec );
  showTrails.Save( wf, sec );
  showVoxels.Save( wf, sec );
  showScreenshots.Save( wf, sec );
  pCamOn.Save( wf, sec );
*/  
}




void Canvas::drawView(uint view)
{
  //Enable the following to debug camera model
  //	if( loaded_texture == true && pCamOn == true )
  //		return;
//  int x = viewOptions[view].viewport[0]
//  int y = viewOptions[view].viewport[1]
  int w = viewOptions[view].viewport[2];
  int h = viewOptions[view].viewport[3];
  bool showVideo = viewOptions[view].showOption(this->showVideo);

  viewOptions[view].SetGlViewport();

   glMatrixMode (GL_PROJECTION);
   glLoadIdentity ();
   glOrtho( 0,100, 0,100, -1, 1 );
   glMatrixMode (GL_MODELVIEW);
   glLoadIdentity();

   glDisable( GL_DEPTH_TEST );
   glColor3f( 0.5, 0.5, 0.5 );
   glRecti( 0,0,100,100 );
   glEnable( GL_DEPTH_TEST );

  //  if (!valid() )
    { 
      if( ! init_done )
        InitGl();
		
      if( viewOptions[view].showOption(pCamOn) == true )
		  {
			 perspective_camera_vector[view].setAspect( static_cast< float >( w ) / static_cast< float >( h ) );
			 perspective_camera_vector[view].SetProjection();
			 current_camera_vector[view] = &perspective_camera_vector[view];
		  } 
      else 
		  {
			 stg_bounds3d_t extent = world->GetExtent();
			 camera_vector[view].SetProjection( w, h, extent.y.min, extent.y.max );
			 current_camera_vector[view] = &camera_vector[view];
		  }
		
///      glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
      glClear( GL_DEPTH_BUFFER_BIT );
    }           

  //Follow the selected robot	
  if( viewOptions[view].last_selection && viewOptions[view].showOption(this->showFollow) )
    {
      Pose gpose = viewOptions[view].last_selection->GetGlobalPose();
      if( viewOptions[view].showOption(pCamOn) == true )
		  {
			 perspective_camera_vector[view].setPose( gpose.x, gpose.y, 0.2 );
			 perspective_camera_vector[view].setYaw( rtod( gpose.a ) - 90.0 );
		  } 
      else 
		  {
			 camera_vector[view].setPose( gpose.x, gpose.y );
		  }
    }

  current_camera_vector[view]->Draw();

if (showVideo) {

/* rie  
setTsaiCameraView(
-2.297800,
0.277710,
-3.129900,
0.923640,
-0.434690,
3.788700,
7.429700,
1.018600,
2304.000000,
3072.000000,
h,
w,
0.002337,
0.002309,
1562.611100,
929.906800);
*/

//july
setTsaiCameraView(
2.352820,
0.112408,
0.022301,
-1.825233,
1.917165,
0.893178,
4.852419,
0.967231,
960.000000,
1296.000000,
552,
706,
0.004390,
0.004450,
596.276506,
477.503208);

/*
setTsaiCameraView(
2.437346,
0.096427,
0.025700,
-1.829398,
2.005563,
0.768883,
4.016948,
1.040883,
960.000000,
1296.000000,
480,
640,
0.004390,
0.004450,
613.020049,
497.718435);
*/
}


  FOR_EACH(it, world->option_table) {
//    printf("%d %s %s\n", view, (*it)->name().c_str(), (*it)->val() ? "T" : "F");
    (*it)->set(this,view);
   }

   renderFrame(view);

if (viewOptions.size() > 1) {   
   glMatrixMode (GL_PROJECTION);
   glLoadIdentity ();
   glOrtho( 0,100, 0,100, -1, 1 );
   glMatrixMode (GL_MODELVIEW);
   glLoadIdentity();

   glDisable( GL_DEPTH_TEST );
   if (view == selected_view) {
    glLineWidth(5);
    glColor3f( 1.0, 0.0, 0.0 );
   } else {
    glLineWidth(3);
    glColor3f( 0.0, 1.0, 0.0 );
   }

   glPolygonMode(GL_FRONT, GL_LINE);
   glRecti( 0,0,100,100 );
   glEnable( GL_DEPTH_TEST );
   glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
   glLineWidth(1);
}

}


void Canvas::draw()
{
if (!valid()) {
  for(uint i=0;i<viewOptions.size();i++) {
    viewOptions[i].SetViewportParams(w(),h());
  }
}
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  for(uint i=0;i<viewOptions.size();i++) {
      drawView(i);
  }
  viewOptions[selected_view].SetGlViewport();
}

void Canvas::resize(int X,int Y,int W,int H) 
{
  Fl_Gl_Window::resize(X,Y,W,H);
  for(uint i=0;i<viewOptions.size();i++) {
    viewOptions[i].SetViewportParams(W,H);
  }
/*
setTsaiCameraView(
-2.297800,
0.277710,
-3.129900,
0.923640,
-0.434690,
3.788700,
7.429700,
1.018600,
2304.000000,
3072.000000,
H,
W,
0.002337,
0.002309,
1562.611100,
929.906800);
*/
  
//  FixViewport(W,H);
  invalidate();
}

//TODO: should be in Canvas
unsigned char ctx[4200000];
unsigned char ctx2[4200000];

static void *lock(void *data, void **p_pixels)
{
//  std::cout << "lock" << std::endl;
  *p_pixels = data;
  return NULL; /* picture identifier, not needed here */
}

static void unlock(void *data, void *id, void *const *p_pixels)
{
//  std::cout << "unlock" << std::endl;
}

static void display(void *data, void *id)
{
//  std::cout << "display" << std::endl;
    (void) data;
    assert(id == NULL);
}


void Canvas::displayVideo() {
   static  GLuint texName = -1;

   //use orthogonal projection without any zoom
   glMatrixMode (GL_PROJECTION);
   glPushMatrix(); //save old projection
   glLoadIdentity ();
   glOrtho( -1,1, -1,1, -1, 1 );
   glMatrixMode (GL_MODELVIEW);
   glPushMatrix();
   
   glLoadIdentity();
   glDisable( GL_DEPTH_TEST );

   //create room for texture
   if (texName == (GLuint)-1) {
      glGenTextures(1, &texName);
   }
   glBindTexture(GL_TEXTURE_2D, texName);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

   unsigned char* pixels = (unsigned char*)(ctx);
   distort->transform(ctx2, ctx);
   glTexImage2D ( GL_TEXTURE_2D, 0, GL_RGB, 2048, 1024, 0,
                      GL_BGR, GL_UNSIGNED_BYTE, 0);
       glTexSubImage2D ( GL_TEXTURE_2D, 0, 0, 0,VIDEOWIDTH ,VIDEOHEIGHT, 
                      GL_BGR, GL_UNSIGNED_BYTE, ctx2);

//   glTexImage2D ( GL_TEXTURE_2D, 0, GL_RGB, VIDEOWIDTH,VIDEOHEIGHT, 0,
//                 GL_BGR, GL_UNSIGNED_BYTE, pixels);

   glEnable(GL_TEXTURE_2D);
   glColor3f( 1.0, 1.0, 1.0 );

   glBegin(GL_QUADS);
     glTexCoord2f(0.0f, VIDEOHEIGHT/1024.0f); glVertex3f(-1.0f, -1.0f,  0.0f); 
     glTexCoord2f(VIDEOWIDTH/2048.0f, VIDEOHEIGHT/1024.0f); glVertex3f( 1.0f, -1.0f,  0.0f); 
     glTexCoord2f(VIDEOWIDTH/2048.0f, 0.0f); glVertex3f( 1.0f,  1.0f,  0.0f); 
     glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,  1.0f,  0.0f); 

//      glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f, -1.0f,  0.0f); 
//      glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f, -1.0f,  0.0f); 
//      glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,  1.0f,  0.0f); 
//      glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,  1.0f,  0.0f); 

   glEnd();

   glDisable(GL_TEXTURE_2D);
//   glEnable(GL_POLYGON_OFFSET_FILL);
//   glPolygonOffset(2.0, 2.0);

   glEnable( GL_DEPTH_TEST );


   //restore camera projection
   glPopMatrix();
   glMatrixMode (GL_PROJECTION);
   glPopMatrix();
   glMatrixMode (GL_MODELVIEW);
}




void Canvas::videoCb(Fl_Widget* w, void* p )
{
  static bool init = false;
  static libvlc_instance_t *libvlc;
  static libvlc_media_t *m;
  static libvlc_media_player_t *mp = 0;
  char const *vlc_argv[] =  {
     "--no-audio", /* skip any audio track */
     "--no-xlib", /* tell VLC to not use Xlib */
     "--rtsp-caching=0", 
  };
  int vlc_argc = sizeof(vlc_argv) / sizeof(*vlc_argv);


  std::cout << "VIDEO CB" << std::endl;
  Canvas* canvas = static_cast<Canvas*>( w );
  
  if (!init) {
    libvlc = libvlc_new(vlc_argc, vlc_argv);
    m = libvlc_media_new_path(libvlc, canvas->videoname.c_str());
    mp = libvlc_media_player_new_from_media(m);
    libvlc_media_release(m);

    libvlc_video_set_callbacks(mp, lock, unlock, display, &ctx);
    libvlc_video_set_format(mp, "RV24", VIDEOWIDTH, VIDEOHEIGHT, VIDEOWIDTH*3);
    canvas->distort = new imr::CRadialDistortion(VIDEOWIDTH, VIDEOHEIGHT,3,VIDEOWIDTH/2, VIDEOHEIGHT/2, 0.000000260);
    init = true;
  }

  
   if ( canvas->showVideo ) { // video on
      std::cout << "VIDEO START" << std::endl;
      libvlc_media_player_play(mp);
   }
   else {
      libvlc_media_player_stop(mp);
      std::cout << "VIDEO DONE" << std::endl;
      //TODO: release vlc
//    libvlc_media_player_release(mp);
//    libvlc_release(libvlc);

   }
   canvas->invalidate();
}
