
#include "stage.hh"
#include "option.hh"
#include "radial_distortion.hh"
#include <map>
#include <stack>

namespace Stg
{
  class Canvas : public Fl_Gl_Window
  {
	 friend class WorldGui; // allow access to private members
	 friend class Model;
    friend class Option;
    friend class OptionsDlg;
  
  private:
    class ViewOptions
    {
      public:
        GLint viewport[4];
        float view_position[4]; // position read from config file
        bool selectedModel;
        std::list<Model*> selected_models;
        Model* last_selection; ///< the most recently selected model (even if it is now unselected).
        std::map<std::string,bool> show;
        ViewOptions(Canvas *canvas);
        void LoadPosition(Worldfile* wf, int sec);
        void SetViewportParams(GLint w, GLint h);
        void SetGlViewport();
        void addShowOption(Option &opt);
        bool setShowOption(std::string name, bool value);
        bool invertShowOption(std::string name);
        bool showOption(Option &opt);
    };
	 
	 class GlColorStack
	 {
	 public:
		GlColorStack() : colorstack() {}
		~GlColorStack() {}
		
		void Push( double r, double g, double b, double a=1.0 )
		{
		  Push( Color(r,g,b,a) );
		}
		
		void Push( Color col )
		{
		  colorstack.push( col );		
		  glColor4f( col.r, col.g, col.b, col.a );
		}
		
		void Pop()
		{
		  if( colorstack.size() < 1 )
			 PRINT_WARN1( "Attempted to ColorStack.Pop() but ColorStack %p is empty",
							  this );
		  else
			 {
				Color& old = colorstack.top();
				colorstack.pop();
				glColor4f( old.r,  old.g,  old.b,  old.a ); 
			 }
		}
		
		unsigned int Length()
		{ return colorstack.size(); }


	 private:
		std::stack<Color> colorstack;
	 } colorstack;
	 
	 std::list<Model*> models_sorted;
  

	 std::vector<Camera*> current_camera_vector;
	 std::vector<OrthoCamera> camera_vector;
	 std::vector<PerspectiveCamera> perspective_camera_vector;

	 bool dirty_buffer;
	 Worldfile* wf;

	 int startx, starty;
	 bool clicked_empty_space;
	 int empty_space_startx, empty_space_starty;
	 ///(even if it is now unselected).
	 GLuint texture;
	 std::string videoname;
	 stg_msec_t interval; // window refresh interval in ms

         imr::CRadialDistortion* distort;
	 void RecordRay( double x1, double y1, double x2, double y2 );
	 void DrawRays();
	 void ClearRays();
	 void DrawGlobalGrid(uint view);
  
	 void AddModel( Model* mod );
	 void RemoveModel( Model* mod );
	 Option //showBlinken,
		showBBoxes,
		showBlocks, 
		showBlur,
		showClock, 
		showData, 
		showFlags,
		showFollow,
		showFootprints, 
		showGrid, 
		showOccupancy, 
		showScreenshots,
		showStatus,
		showTrailArrows, 
		showTrailRise, 
		showTrails, 
		showVoxels,
		pCamOn,
      showVideo,
		visualizeAll;
    std::vector<ViewOptions> viewOptions;
    void displayVideo();

  
  public:
	 Canvas( WorldGui* world, int x, int y, int width, int height);
	 ~Canvas();
  
	 bool graphics;
	 WorldGui* world;
	 unsigned long frames_rendered_count;
	 int screenshot_frame_skip;
	 uint selected_view;
    std::map< std::string, Option* > _custom_options;
  

	 void Screenshot(std::string name="");
	 void InitGl();
	 void createMenuItems( Fl_Menu_Bar* menu, std::string path );
  
	 void FixViewport(int W,int H);
	 void DrawFloor(); //simpler floor compared to grid
	 void DrawBlocks();
	 void DrawBoundingBoxes();
	 void resetCamera(uint view);
    void changeView(uint view);
    int getView(int mx, int my);

    
	 virtual void renderFrame(uint view);
	 virtual void drawView(uint view);
	 virtual void draw();
	 virtual int handle( int event );
	 void resize(int X,int Y,int W,int H);

	 void CanvasToWorld( int px, int py, 
					double *wx, double *wy, double* wz, uint view );

	 Model* getModel( int x, int y );
	 bool selected( Model* mod );
	 void select( Model* mod );
	 void unSelect( Model* mod );
	 void unSelectAll();

	 inline void setDirtyBuffer( void ) { dirty_buffer = true; }
	 inline bool dirtyBuffer( void ) const { return dirty_buffer; }
	
	 void PushColor( Color col )
	 { colorstack.Push( col.r, col.g, col.b, 1.0/*0.4*/); }

	 void PushColor( double r, double g, double b, double a )
	 { colorstack.Push( r,g,b,a ); }

	 void PopColor(){ colorstack.Pop(); } 
  
	 void InvertView( uint32_t invertflags );

	 bool VisualizeAll(){ return ! visualizeAll; }
  
	 static void TimerCallback( Canvas* canvas );
	 static void perspectiveCb( Fl_Widget* w, void* p );
    static void videoCb(Fl_Widget* w, void* p );

  
	 void EnterScreenCS();
	 void LeaveScreenCS();

	 void Load( Worldfile* wf, int section );
	 void Save( Worldfile* wf, int section );
  };

} // namespace Stg
