// file: model_minedetector.cc
// Author: Miroslav Kulich
// Date: 17 June 2010
//
///////////////////////////////////////////////////////////////////////////


#include <sys/time.h>

#include "stage.hh"
#include "worldfile.hh"
#include "option.hh"
using namespace Stg;

// DEFAULT PARAMETERS FOR LASER MODEL
static const bool DEFAULT_FILLED = true;
static const stg_watts_t DEFAULT_WATTS = 17.5;
static const Size DEFAULT_SIZE( 0.15, 0.15, 0.2 );
static const stg_meters_t DEFAULT_MAXRANGE = 2.0;
static const stg_radians_t DEFAULT_FOV = 2*M_PI;
static const char* DEFAULT_COLOR = "green";

Option ModelMineDetector::showMineDetector( "Mine detector", "show_minedetector", "", true, NULL );
Option ModelMineDetector::showFov( "Mine detector FOV", "show_minedetector_fov", "", true, NULL );
Option ModelMineDetector::showArea( "Mine detector FOV area", "show_minedetector_area", "", false, NULL );
Option ModelMineDetector::showMines( "Mines", "show_mines", "", false, NULL );

/**
   @ingroup model
   @defgroup model_minedetector Mine detector model
   The mine detector model simulator

   API: Stg::ModelMineDetector

   <h2>Worldfile properties</h2>

   @par Summary and default values

//TODO:
   @verbatim
   minedetector
   (
   # laser properties
   samples 180
   range_max 8.0
   fov 3.14159
   resolution 1

   # model properties
   size [ 0.15 0.15 0.2 ]
   color "blue"
   )
   @endverbatim

   @par Details
 
   - samples <int>\n
   the number of laser samples per scan
   - range_max <float>\n
   the maximum range reported by the scanner, in meters. The scanner will not detect objects beyond this range.
   - fov <float>\n
   the angular field of view of the scanner, in radians. 
   - resolution <int>\n
   Only calculate the true range of every nth laser sample. The missing samples are filled in with a linear interpolation. Generally it would be better to use fewer samples, but some (poorly implemented!) programs expect a fixed number of samples. Setting this number > 1 allows you to reduce the amount of computation required for your fixed-size laser vector.
*/
  

ModelMineDetector::ModelMineDetector( World* world,
												Model* parent,
												const std::string& type ) :
	Model( world, parent, type ), 
	samples(),
	range_max( DEFAULT_MAXRANGE ),
	fov( DEFAULT_FOV )
{
  puts("\nMINE DETECTOR\n");
  PRINT_DEBUG2( "Constructing ModelMineDetector %d (%s)\n",
								id, typestr );
  
  SetGeom( Geom( Pose(0,0,0,0), DEFAULT_SIZE) );
  
  // assert that Update() is reentrant for this derived model
  thread_safe = true;

  SetColor( Color(DEFAULT_COLOR));


  // start consuming power
  watts = DEFAULT_WATTS;
  RegisterOption( &showMineDetector );
  RegisterOption( &showFov );
  RegisterOption( &showArea );
  RegisterOption( &showMines );
}


ModelMineDetector::~ModelMineDetector( void )
{
}

void ModelMineDetector::Load( void )
{  
  range_max = wf->ReadLength( wf_entity, "range_max", range_max );
  fov       = wf->ReadAngle( wf_entity, "fov",  fov );
  Model::Load();
}

ModelMineDetector::Config ModelMineDetector::GetConfig()
{ 
  Config cfg;
  cfg.range_bounds.max = range_max;
  cfg.fov = fov;
  return cfg;
}

void ModelMineDetector::SetConfig( Config& cfg )
{ 
  range_max = cfg.range_bounds.max;
  fov = cfg.fov;
}

static bool raytrace_match( Model* hit, Model* finder, const void* dummy )
{
  (void)dummy; // avoid warning about unused var

  // Ignore the model that's looking and things that are invisible to
  return( (hit != finder) && (hit->vis.obstacle_return > 0 ) &&  (hit->vis.mine_return == 0 ) );
}

void ModelMineDetector::Update( void )
{     
  Pose rr( geom.pose );
  rr = LocalToGlobal(rr);
  
  Pose rayorg( geom.pose );
  rayorg.z += geom.size.z/2.0;
  rayorg = LocalToGlobal(rayorg);
  rayorg.a = 0; // dummy
  Ray ray( this, rayorg, 100/*range_max*/, raytrace_match, NULL, true );

  samples.clear();
  FOR_EACH( it, world->GetChildren() ) {
        if ( (*it)->vis.mine_return ) {  //is it  a bomb?
          Pose p = (*it)->GetPose();
          double dy = p.y-ray.origin.y;
          double dx = p.x-ray.origin.x;
          double dd = sqrt(dx*dx+dy*dy);
          if ( dd<=range_max ) { // in sensor range?
            ray.origin.a = atan2(p.y-ray.origin.y,p.x-ray.origin.x); // in radians
            stg_raytrace_result_t r( world->Raytrace( ray ) );
            if ( dd < r.range ) { // is it visible? (no obstacle between sensor and the bomb
              Sample bomb;
              bomb.angle = ray.origin.a-rr.a;
              if (bomb.angle < -M_PI) {
                bomb.angle += 2*M_PI;
              } else
              if (bomb.angle > M_PI) {
                bomb.angle -= 2*M_PI;
              }
              if (fabs(bomb.angle) <= fov/2) {
                bomb.range = dd;
                bomb.pose = p;
                samples.push_back(bomb);
              }
            }
//            printf("%s  (%lf %lf %lf) %lf",(*it)->Token(),p.x,p.y,p.z,dd);
//            printf("    hit %s: %s (%lf %lf %lf)  range:%lf\n",dd<r.range? "YES" : "no", r.mod ? r.mod->Token() : "xxx", r.pose.x, r.pose.y, r.pose.z, r.range);
          }
        }
      }
  
  Model::Update();
}


void ModelMineDetector::Startup(  void )
{ 
  Model::Startup();
  PRINT_DEBUG( "minedetector startup" );
}

void ModelMineDetector::Shutdown( void )
{ 
  PRINT_DEBUG( "minedetector shutdown" );
  Model::Shutdown();
}

void ModelMineDetector::Print( char* prefix )
{
  Model::Print( prefix );

  printf( "\tRanges[ " );
//  for( unsigned int i=0; i<sample_count; i++ )
//		printf( "%.2f ", samples[i].range );
  puts( " ]" );

}


const std::vector<ModelMineDetector::Sample>& ModelMineDetector::GetSamples()
{ 
  return samples;
}

bool ModelMineDetector::ShowMines() {
  return showMines.val();
};


void ModelMineDetector::DataVisualize( Camera* cam )
{
  (void)cam; // avoid warning about unused var
    

  glPushMatrix();
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  
  glTranslatef( 0,0, GetGeom().size.z/2.0 ); // shoot the laser beam out at the right height
  
  // pack the laser hit points into a vertex array for fast rendering
  static std::vector<GLfloat> pts;

  
  PushColor( 1, 0, 0, 0.5 );
  glDepthMask( GL_FALSE );
  glPointSize( 2 );
/*  
  for( unsigned int s(0); s<sample_count; s++ )
		{
			double ray_angle = (s * (laser->fov / (sample_count-1))) - laser->fov/2.0;
			pts[2*s+2] = (float)(samples[s].range * cos(ray_angle) );
			pts[2*s+3] = (float)(samples[s].range * sin(ray_angle) );
		
			// if the sample is unusually bright, draw a little blob
			if( samples[s].reflectance > 0 )
				{
					glBegin( GL_POINTS );
					glVertex2f( pts[2*s+2], pts[2*s+3] );
					glEnd();
				}			 
		}
  
  glVertexPointer( 2, GL_FLOAT, 0, &pts[0] );       
  
  laser->PopColor();
*/

/*  
  if( showArea )
		{
			// draw the filled polygon in transparent blue
			laser->PushColor( 0, 0, 1, 0.1 );		
			glDrawArrays( GL_POLYGON, 0, sample_count+1 );
			laser->PopColor();  
		}
*/  
  glDepthMask( GL_TRUE );

  if ( showMineDetector ) {
    PushColor( 1, 0, 0, 1 );
    glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
//    glPointSize( 20 );
    FOR_EACH(s,samples) {
      double x = cos(s->angle)*s->range;
      double y = sin(s->angle)*s->range;
    glBegin(GL_POLYGON);
      for(int i=0;i<30;i++) {
        glVertex3f( x+0.11*cos(2*M_PI*(i/30.0)), y+0.11*sin(2*M_PI*(i/30.0)), 0.1 );
      }
    glEnd();
    }      
    PopColor();
  }
  
  if( showFov || showArea)
		{
        unsigned int resolution = 30;
        unsigned int k = 0;
        pts.resize( 2 * (resolution+1) );
         if (fabs(fov-2*M_PI) > 0.01) {
           pts[0] = 0;
           pts[1] = 0;
           k = 1;
         } 


        for( unsigned int s(0); s<resolution; s++ )//TODO: magic constant
				{
					double ray_angle((s * (fov / (resolution-1))) - fov/2.0);
					pts[2*(s+k)] = (float)(range_max * cos(ray_angle) );
					pts[2*(s+k)+1] = (float)(range_max * sin(ray_angle) );
				}
        
         if (showFov) {
          glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
          PushColor( 0, 1, 0, 1 );
          glVertexPointer( 2, GL_FLOAT, 0, &pts[0] );
          glDrawArrays( GL_POLYGON, 0, resolution +k-1);
          PopColor();
         }
         if (showArea) {
          glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
          PushColor( 0, 1, 0, 0.1 );
          glVertexPointer( 2, GL_FLOAT, 0, &pts[0] );
          glDrawArrays( GL_POLYGON, 0, resolution +k-1);
          PopColor();
         }
         
		}

  glPopMatrix();
}

