#include "stage.hh"
#include "model_occgrid.hh"
#include "canvas.hh"

using namespace Stg;

   bool operator< (const Pixel &px1, const Pixel &px2) {
      return px1.x < px2.x || (px1.x == px2.x && px1.y < px2.y);
   }

ModelOccGrid::ModelOccGrid( World* world, Model* parent, const std::string& type ) :
Model( world, parent, type ), 
rows (0),
columns (0),
cells(0),
vis( world )
{  
   PRINT_DEBUG2( "Constructing ModelOccGrid %d (%s)\n",id, typestr );
   
//   SetGeom( Geom( Pose(0,0,0,0), DEFAULT_SIZE) );
   
  // assert that Update() is reentrant for this derived model
   thread_safe = true;
   
   AddVisualizer( &vis, true );  
}


ModelOccGrid::~ModelOccGrid( void ) {
   if (cells) {
      for(uint i = 0;i<rows;i++) {
         delete [] cells[i];
      }
      delete [] cells;
   }
}




void ModelOccGrid::Load( void ) {  
   cell_size = wf->ReadFloat( wf_entity, "cell_size", 10.0 )/100.0; // convert to meters
   alpha_color = wf->ReadFloat( wf_entity, "alpha_color", 1.0 ); // convert to meters
   Model::Load();
}

ModelOccGrid::Config ModelOccGrid::GetConfig() { 
   Config cfg;
   cfg.cell_size = cell_size;
   cfg.alpha_color = alpha_color;
   return cfg;
}

void ModelOccGrid::SetConfig( Config& cfg ) {
  cell_size = cfg.cell_size;
  alpha_color = cfg.alpha_color;
}


void ModelOccGrid::SampleConfig() {
}

void ModelOccGrid::Update( void )
{     
   Model::Update();
}



void ModelOccGrid::Startup(  void )
{ 
   Model::Startup();
   PRINT_DEBUG( "occgrid startup" );
}

void ModelOccGrid::Shutdown( void )
{ 
   PRINT_DEBUG( "occgrid shutdown" );
   Model::Shutdown();
}

void ModelOccGrid::Print( char* prefix )
{
   Model::Print( prefix );
}

// VIS -------------------------------------------------------------------

ModelOccGrid::Vis::Vis( World* world ) 
: Visualizer( "OccGrid", "occgrid_vis" )
{
}

void ModelOccGrid::Vis::DrawPixels(Model* mod, ColoredPixels pixels) {
   ModelOccGrid *occgrid = (ModelOccGrid*) mod;
   mod->PushColor( pixels.color );
   FOR_EACH(cp,pixels.pixels) {
      glRectf(cp->x*occgrid->cell_size,cp->y*occgrid->cell_size,(cp->x+1.0)*occgrid->cell_size,(cp->y+1.0)*occgrid->cell_size);
   }
   mod->PopColor();
}

void ModelOccGrid::Vis::SetPixelsDepth(Model* mod, ColoredPixels pixels) {
   ModelOccGrid *occgrid = (ModelOccGrid*) mod;
   std::map<Pixel,DepthColor>::iterator it; 
   FOR_EACH(cp,pixels.pixels) {
      occgrid->cells[cp->y][cp->x].draw = false; 
      it = occgrid->depths.find(*cp);
      if (it == occgrid->depths.end()) {
         occgrid->depths[*cp] = DepthColor(pixels.depth,pixels.color);
      }  else {
         if (it->second.depth <= pixels.depth) {
            it->second.depth = pixels.depth;
            it->second.color = pixels.color;
         }
      }
   }
}



void ModelOccGrid::Vis::Visualize( Model* mod, Camera* cam ) 
{
  
   //   glMatrixMode (GL_PROJECTION);
   glPushMatrix();
   /*
   glLoadIdentity ();
   glOrtho( 0, 200,0, 200, -100, 100 );	
   glMatrixMode (GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();
   glDisable( GL_DEPTH_TEST );
   */
   Gl::pose_inverse_shift( mod->GetGlobalPose() );
   glPolygonMode( GL_FRONT, GL_FILL );
   //   glTranslatef(0,0,-1.0);
   ModelOccGrid *occgrid = (ModelOccGrid*) mod;
   
   occgrid->depths.clear();
   FOR_EACH(p,occgrid->pixels) {
      SetPixelsDepth(mod, p->second);
   }
   
   float color;
   for(uint i=0;i<occgrid->rows;i++) {
      for(uint j=0;j<occgrid->columns;j++) {
         if (occgrid->cells[i][j].draw) {
            color = occgrid->cells[i][j].value/255.0;
            mod->PushColor( color,color,color,occgrid->alpha_color );
            glRectf(j*occgrid->cell_size,i*occgrid->cell_size,(j+1.0)*occgrid->cell_size,(i+1.0)*occgrid->cell_size);
            mod->PopColor();
         }
      }
   }

   FOR_EACH(cp,occgrid->depths) {
      mod->PushColor( cp->second.color );
      glRectf(cp->first.x*occgrid->cell_size,cp->first.y*occgrid->cell_size,(cp->first.x+1.0)*occgrid->cell_size,(cp->first.y+1.0)*occgrid->cell_size);
      mod->PopColor();
   }

   //   glTranslatef(0,0,0.2);
   
   //   glEnable( GL_DEPTH_TEST );
   glPopMatrix();
      //restore camera projection
   //   glMatrixMode (GL_PROJECTION);
   /*
   glPopMatrix();
   glMatrixMode (GL_MODELVIEW);
*/
}


void ModelOccGrid::ReallocCells(uint icols, uint irows) {
   if (cells) {
      for(uint i = 0;i<rows;i++) {
         delete [] cells[i];
      }
      delete [] cells;
   }
   
   rows =irows;
   columns = icols;
   cells = new GCell* [rows];
   for(uint i = 0;i<rows;i++) {
      cells[i] = new GCell[columns]; 
   }
}

void ModelOccGrid::SetCells(uchar *icells) {
   //   uint  rowsize = rows*sizeof(uchar);
   for(uint i=0,k=0;i<rows;i++) {
      for(uint j=0;j<columns;j++,k++) {
         cells[i][j].value = icells[k];
         cells[i][j].draw = true;
      }
   }
   
   //   for(uint i=0, k=0;i<rows;i++,k+=rowsize) {
   //      std::copy(&icells[k],&icells[k+rowsize],&cells[i][0]);
   //}
}

void ModelOccGrid::AddPixels(const std::string name,  const Color &color, const uint8_t type, const uint8_t depth, const std::vector<Pixel> &pixels) {
   PRINT_DEBUG2("Got pixels:  %s %d\n",name.c_str(), pixels.size());
   if (pixels.empty()) {
      this->pixels.erase(name);
   } else {
      this->pixels[name] = ColoredPixels(color,type,depth,pixels); 
   }
}
