/*
 *  Player - One Hell of a Robot Server
 *  Copyright (C) 2004, 2005 Richard Vaughan
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * Desc: A plugin driver for Player that gives access to Stage devices.
 * Author: Miroslav Kulich
 * Date: 21 June 2010
 */

// DOCUMENTATION

/**
	@addtogroup player
	@par Mine detector interface
	- PLAYER_MINEDETECTOR_DATA_SCAN
	- PLAYER_MINEDETECTOR_REQ_GET_GEOM
*/

/* TODO
	- PLAYER_MINEDETECTOR_REQ_SET_FOV
	- PLAYER_MINEDETECTOR_REQ_GET_FOV
*/

// CODE

#include "p_driver.h"
using namespace Stg;


InterfaceMineDetector::InterfaceMineDetector(  player_devaddr_t addr,
													StgDriver* driver,
													ConfigFile* cf,
													int section )
  : InterfaceModel( addr, driver, cf, section, "minedetector" )
{
}

void InterfaceMineDetector::Publish( void )
{
  player_minedetector_data_t pdata;
  memset( &pdata, 0, sizeof(pdata) );
  
	const std::vector<ModelMineDetector::Sample>& bombs =
		((ModelMineDetector*)mod)->GetSamples();
	
	pdata.mines_count = bombs.size();

	if( pdata.mines_count > 0 )
    {
	   pdata.mines = new player_minedetector_item_t[pdata.mines_count];
			
      for( unsigned int i=0; i<pdata.mines_count; i++ )
		{
			pdata.mines[i].angle = bombs[i].angle;
         pdata.mines[i].range = bombs[i].range;
		}
    }
	
  // publish this data
  this->driver->Publish( this->addr,
												 PLAYER_MSGTYPE_DATA,
												 PLAYER_MINEDETECTOR_DATA_SCAN,
												 &pdata, sizeof(pdata), NULL);
	
  if ( pdata.mines )
		delete [] pdata.mines;
}

int InterfaceMineDetector::ProcessMessage(QueuePointer& resp_queue,
												  player_msghdr_t* hdr,
												  void* data )
{
  //printf("got fiducial request\n");

  // Is it a request to get the geometry?
  if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_REQ,
                           PLAYER_MINEDETECTOR_REQ_GET_GEOM,
                           this->addr))
    {
      Geom geom = mod->GetGeom();
      Pose pose = mod->GetPose();

      // fill in the geometry data formatted player-like
      player_laser_geom_t pgeom;
      pgeom.pose.px = pose.x;
      pgeom.pose.py = pose.y;
      pgeom.pose.pz = pose.z;
      pgeom.pose.proll = 0.0;
      pgeom.pose.ppitch = 0.0;
      pgeom.pose.pyaw = pose.a;
      pgeom.size.sl = geom.size.x;
      pgeom.size.sw = geom.size.y;

      this->driver->Publish(this->addr, resp_queue,
									 PLAYER_MSGTYPE_RESP_ACK,
									 PLAYER_MINEDETECTOR_REQ_GET_GEOM,
									 (void*)&pgeom, sizeof(pgeom), NULL);
      return(0);
    }
  /*    case PLAYER_MINEDETECTOR_SET_FOV:

		  if( len == sizeof(player_fiducial_fov_t) )
		  {
		  player_fiducial_fov_t* pfov = (player_fiducial_fov_t*)src;

		  // convert from player to stage FOV packets
		  stg_fiducial_config_t setcfg;
		  memset( &setcfg, 0, sizeof(setcfg) );
		  setcfg.min_range = (uint16_t)ntohs(pfov->min_range) / 1000.0;
		  setcfg.max_range_id = (uint16_t)ntohs(pfov->max_range) / 1000.0;
		  setcfg.max_range_anon = setcfg.max_range_id;
		  setcfg.fov = DTOR((uint16_t)ntohs(pfov->view_angle));

		  //printf( "setting fiducial FOV to min %f max %f fov %f\n",
		  //  setcfg.min_range, setcfg.max_range_anon, setcfg.fov );

		  //stg_model_set_config( this->mod, &setcfg, sizeof(setcfg));
		  stg_model_set_property( this->mod, "fiducial_cfg",
		  &setcfg, sizeof(setcfg));
		  }
		  else
		  PRINT_ERR2("Incorrect packet size setting fiducial FOV (%d/%d)",
		  (int)len, (int)sizeof(player_fiducial_fov_t) );

		  // deliberate no-break - SET_FOV needs the current FOV as a reply

		  case PLAYER_MINEDETECTOR_GET_FOV:
		  {
		  stg_fiducial_config_t *cfg = (stg_fiducial_config_t*)
		  stg_model_get_property_fixed( this->mod, "fiducial_cfg", sizeof(stg_fiducial_config_t));
		  assert(cfg);

		  // fill in the geometry data formatted player-like
		  player_fiducial_fov_t pfov;
		  pfov.min_range = htons((uint16_t)(1000.0 * cfg->min_range));
		  pfov.max_range = htons((uint16_t)(1000.0 * cfg->max_range_anon));
		  pfov.view_angle = htons((uint16_t)RTOD(cfg->fov));

		  if( this->driver->PutReply(  this->id, client, PLAYER_MSGTYPE_RESP_ACK,
		  &pfov, sizeof(pfov), NULL ) != 0 )
		  DRIVER_ERROR("PutReply() failed for "
		  "PLAYER_MINEDETECTOR_GET_FOV or PLAYER_MINEDETECTOR_SET_FOV");
		  }
		  break;
  */

  else
    {
      // Don't know how to handle this message.
      PRINT_WARN2( "stg_minedetector doesn't support msg with type/subtype %d/%d",
						 hdr->type, hdr->subtype);
      return(-1);
    }

  return 0;
}

